import numpy as np
import abipy.core.abinit_units as abu

from numpy import fft
from scipy import signal
try:
    from scipy.integrate import simpson as simps
except ImportError:
    from scipy.integrate import simps

from pymatgen.io.phonopy import get_pmg_structure
from abipy.tools.plotting import get_ax_fig_plt,add_fig_kwargs
from abipy.embedding.utils_ifc import clean_structure

class Lineshape():
    """

    Object representing a luminescent lineshape, following a multi-phonon mode model (multiD-CCM).
    For 1D-CCM, use plot_lineshape_1D_zero_temp() function of the abipy/lumi/deltaSCF module.

    For equations, notations and formalism, please refer to :
     https://doi.org/10.1103/PhysRevB.96.125132
     https://doi.org/10.1002/adom.202100649
     https://pubs.acs.org/doi/full/10.1021/acs.chemmater.3c00537

    In the 1D-CCM, the vibronic peaks are the ones from a fictious phonon mode that connects
    the atomic relaxation between the ground state and excited state.
    Within this model, the global shape (fwhm) is well represented if the total Huang-Rhys
    factor is large enough (gaussian shaped spectrum). However, if vibronic
    peaks are present in the experimental spectrum, this model is not able to correctly
    reproduce them as it assumes a fictious phonon mode.

    In the multiD-CCM, the atomic relaxation is projected along the phonon eigenvectors of the system,
    allowing a phonon-projected decomposition of the relaxation.
    Better agreement with experimental vibronic peaks is expected.
    """

    @classmethod
    def from_phonopy_phonons(cls,E_zpl,phonopy_ph,dSCF_structure,
                             use_forces=True,dSCF_displacements=None,dSCF_forces=None,coords_defect_dSCF=None,tol=0.3):
        r"""
        Different levels of approximations for the phonons and force/displacements:
        See discussion in the supplementary informations of
        https://pubs.acs.org/doi/full/10.1021/acs.chemmater.3c00537, section (1).

        - size_supercell deltaSCF = size_supercell phonons (phonons of the bulk structure or phonons of defect structure).
          Use of the forces or the displacemements is allowed.

        - size_supercell dSCF < size_bulk_supercell phonons (bulk)
          Use of the forces only.

        - size_supercell dSCF < size__defect_supercell phonons (embedding)
          Use of the forces only

        The code first extracts the eigenmodes of the phonopy object.
        Then, it tries performs a structure matching between the phonopy
        structure and the dSCF structure (critical part) in order to put the displacements/forces on the right atoms.

        Args:
            E_zpl: Zero-phonon line energy in eV
            phonopy_ph: Phonopy object containing eigenfrequencies and eigenvectors
            dSCF_structure: Delta_SCF structure
            dSCF_displacements: Dispalcements \Delta R induced by the electronic transition
            dSCF_forces: Dispalcements \Delta F induced by the electronic transition
            coords_defect_dSCF: Main coordinates of the defect in defect structure, if defect complex, can be set to the
                center of mass of the complex
            tol: tolerance in Angstrom applied for the matching between the dSCF structure and phonon structure

        Returns: A lineshape object
        """

        ph_modes = phonopy_ph.get_frequencies_with_eigenvectors(q=[0, 0, 0])
        ph_freq_phonopy, ph_vec_phonopy = ph_modes

        freqs = ph_freq_phonopy * (1/ abu.eV_to_THz)   # THz to eV
        vecs = ph_vec_phonopy.transpose() #

        dSCF_structure=clean_structure(dSCF_structure,coords_defect_dSCF)

        phonon_supercell=get_pmg_structure(phonopy_ph.supercell)
        phonon_supercell=clean_structure(phonon_supercell,coords_defect_dSCF)

        if use_forces==False:
            forces=None
            displacements=get_displacements_on_phonon_supercell(dSCF_supercell=dSCF_structure,
                                                                phonon_supercell=phonon_supercell,
                                                                displacements_dSCF=dSCF_displacements,
                                                                tol=tol)

        if use_forces==True:
            displacements=None
            forces=get_forces_on_phonon_supercell(dSCF_supercell=dSCF_structure,
                                                 phonon_supercell=phonon_supercell,
                                                 forces_dSCF=dSCF_forces,
                                                 tol=tol)

        return cls(E_zpl=E_zpl,
                   ph_eigvec=vecs,
                   ph_eigfreq=freqs,
                   structure=phonon_supercell,
                   use_forces=use_forces,
                   forces=forces,
                   displacements=displacements)


    def __init__(self, E_zpl, ph_eigvec, ph_eigfreq, structure,
                 forces,displacements,use_forces):
        """

        Args:
            E_zpl: Zero-phonon line energy in eV
            ph_eigvec: phonon eigenvectors, shape : (3 * N_atoms, 3 * N_atoms)
            ph_eigfreq: phonon eigenfrequencies, shape : (3 * N_atoms), in eV
            structure: Structure object
            forces : Forces acting on the atoms in the ground state, with atomic positions of the relaxed excited state, in eV/Ang
            displacements : Atomic relaxation induced by the electronic transition, in Ang
            use_forces : True in order to use the forces, False to use the displacements
        """
        self.E_zpl = E_zpl
        self.ph_eigvec = ph_eigvec
        self.ph_eigfreq = ph_eigfreq
        self.structure = structure
        self.use_forces=use_forces
        self.forces = forces
        self.displacements = displacements

    def n_modes(self):
        """
        Number of phonon modes
        """
        return len(self.ph_eigfreq)

    def mass_list(self):
        """
        List of masses of the atoms in the structure, in amu unit
        """
        #
        amu_list = np.zeros(len(self.structure))
        for i, atom in enumerate(self.structure.species):
            amu_list[i] = atom.atomic_mass
        return (amu_list)

    def Delta_Q_nu(self):
        """
        List of Delta_Q_nu of the atoms in the structure, in SI unit
        """
        Q_nu = np.zeros(self.n_modes())

        masses = np.repeat(self.mass_list(), 3) * 1.66053892173E-27 # amu to kg
        ph_eigvector = self.ph_eigvec
        ph_eigfreq= self.ph_eigfreq * (abu.eV_s)  # eV to rad/s

        if self.use_forces==True:
            delta_forces=self.forces.flatten() * ((abu.eV_Ha)*(abu.Ha_J)/(1e-10)) # eV/Angstrom to Newtons (Joules/meter)
            for i in range(self.n_modes()):
                if self.ph_eigfreq[i] < 1e-5 :# discard phonon freq with too low freq (avoid division by nearly 0)
                    Q_nu[i] = 0
                else:
                    Q_nu[i] = (1/ph_eigfreq[i]**2) * np.sum( delta_forces/np.sqrt(masses) * np.real(ph_eigvector[i]) )
                # equation (7) of Alkauskas, A. (2014) New Journal of Physics, 16(7), 073026.
        else:
            displacements = self.displacements.flatten() * (1e-10) # in meters
            for i in range(self.n_modes()):
                Q_nu[i] = np.sum(np.sqrt(masses) * displacements * np.real(ph_eigvector[i]))
                # equation (6) of Alkauskas, A. (2014) New Journal of Physics, 16(7), 073026.`

        Q_nu[0:3]=0 # acoustic modes are set to 0
        return Q_nu


    def S_nu(self):
        """
        Partial Huang-Rhys factors
        """
        omega = (abu.eV_s) * self.ph_eigfreq # eV to [rad/s]
        Delta_Q = self.Delta_Q_nu()
        hbar = abu.hbar_eVs*((abu.eV_Ha)*(abu.Ha_J)) # hbar in SI
        S_nu = omega * Delta_Q ** 2 / (2 * hbar)
        return (S_nu)

    def S_tot(self):
        """
        Total Huang-Rhys factor = sum of the S_nu.
        """
        return (np.sum(self.S_nu()))

    def Delta_Q(self):
        """
        Total Delta_Q
        """
        return (np.sqrt(np.sum(self.Delta_Q_nu() ** 2)))

    def p_nu(self):
        """
        Contribution of each mode, eq. (11) of  https://doi.org/10.1002/adom.202100649
        """
        return (self.Delta_Q_nu() / self.Delta_Q()) ** 2

    def eff_freq_multiD(self):
        """
        Effective coupling frequency, eq. (13) of  https://doi.org/10.1002/adom.202100649
        """
        w = np.sqrt(np.sum(self.p_nu() * self.ph_eigfreq ** 2))
        return (w)

    def S_hbarOmega(self, broadening):
        """
        Return (Energy,spectral decomposition S(hw))

        Args:
            broadening: fwhm of the gaussian broadening in meV
        """
        n_step = 100001
        S = np.zeros(n_step)
        sigma = broadening / (1000 * 2.35482)
        freq_eV = self.ph_eigfreq
        omega = np.linspace(0, 1.1 * max(freq_eV), n_step)
        S_nu = self.S_nu()

        for k in np.arange(self.n_modes()):
            S += S_nu[k] * (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-((omega - freq_eV[k]) ** 2 / (2 * sigma ** 2)))
        return (omega, S)

    #### Generating function ####

    def Bose_einstein(self, T, freq):
        """
        Bose Einstein average occupation number

        Args:
            T: Temperature in K
            freq: frequency in eV
        """
        k_b = abu.kb_eVK  # in eV/K

        if T == 0:
            n = 0
        else:
            n = 1 / (np.exp(freq / (k_b * T)) - 1)

        return n

    def G_t(self, T, S_nu, omega_nu):
        """
        Generation function.
        Eq. (3) of https://pubs.acs.org/doi/full/10.1021/acs.chemmater.3c00537

        Args:
            T: Temperature in K
            S_nu: Parial Huang-Rhys factors
            omega_nu: phonon frequencies
        """
        n_step = 10001
        t = np.linspace(-1e-11, +1e-11, n_step)  # time in the fourier domain

        freq = omega_nu
        freq_SI = freq * (abu.eV_s) # in SI rad/sec

        S=np.zeros(n_step, dtype=complex)
        C_plus=np.zeros(n_step, dtype=complex)
        C_minus=np.zeros(n_step, dtype=complex)

        for i in range(len(S_nu)):
            S +=  S_nu[i] * np.exp(-1j * freq_SI[i] * t)
            C_plus += self.Bose_einstein(T, freq[i]) * S_nu[i] * np.exp(+1j * freq_SI[i] * t)
            C_minus += self.Bose_einstein(T, freq[i]) * S_nu[i] * np.exp(-1j * freq_SI[i] * t)

        index_0 = int((len(t) - 1) / 2)
        C_0=2*C_plus[index_0]
        S_0=S[index_0]

        G_t = np.exp(S-S_0+C_plus+C_minus-2*C_0)

        return (t, G_t)

    def A_hw(self, T, lamb=5, w=1, model='multi-D'):
        """
        Lineshape function
        Eq. (2) of https://pubs.acs.org/doi/full/10.1021/acs.chemmater.3c00537
        Returns (Energy in eV, Lineshape function )

        Args:
            T: Temperature in K
            lamb: Lorentzian broadening applied to the vibronic peaks, in meV
            w: Gaussian broadening applied to the vibronic peaks, in meV
            model: 'multi-D' for full phonon decomposition, 'one-D' for 1D-CCM PL spectrum.
        """
        if model == 'multi-D':
            t, G_t = self.G_t(T, S_nu=self.S_nu(), omega_nu=self.ph_eigfreq)

        elif model == 'one-D':
            t, G_t = self.G_t(T, S_nu=np.array([np.sum(self.S_nu())]), omega_nu=np.array([self.eff_freq_multiD()]))

        n_step = len(t)
        Lambda = abu.eV_s*0.001/(np.pi*2) * lamb  # meV to Hz
        delta_t = t[-1] - t[0]

        fourier = fft.fft(G_t * np.exp(-Lambda * np.abs(t)))
        freq = 2 * np.pi * n_step * fft.fftfreq(G_t.size) / delta_t
        # freq change
        freq_2 = np.zeros(n_step)
        fourier_2 = np.zeros(n_step, dtype=complex)

        freq_2[0:n_step // 2] = freq[n_step // 2 + 1:]
        freq_2[n_step // 2] = freq[0]
        freq_2[n_step // 2 + 1:] = freq[1:n_step // 2 + 1]

        fourier_2[0:n_step // 2] = fourier[n_step // 2 + 1:]
        fourier_2[n_step // 2] = fourier[0]
        fourier_2[n_step // 2 + 1:] = fourier[1:n_step // 2 + 1]

        hbar_eV = abu.hbar_eVs  # in eV*s
        En = hbar_eV * freq_2
        E_x = En + self.E_zpl

        sigma = w / (2.35482 * 1000)
        gaussian = (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-((En) ** 2 / (2 * (sigma) ** 2)))
        A_conv = signal.fftconvolve(np.abs(fourier_2), gaussian, mode='same')

        return (E_x, A_conv)

    def L_hw(self, T=0, lamb=5, w=1, model='multi-D'):
        """
        Normalized Luminescence intensity (area under the curve = 1)
        Eq. (1) of https://pubs.acs.org/doi/full/10.1021/acs.chemmater.3c00537
        Returns (Energy in eV, Luminescence intensity)

        Args:
            T: Temperature in K
            lamb: Lorentzian broadening applied to the vibronic peaks, in meV
            w: Gaussian broadening applied to the vibronic peaks, in meV
            model: 'multi-D' for full phonon decomposition, 'one-D' for 1D-CCM PL spectrum.
        """

        E_x, A = self.A_hw(T, lamb, w, model)
        C = 1 / (simps(A * E_x ** 3, x=E_x))
        I = C * A * E_x ** 3  # intensity prop to energy CUBE
        return E_x, I

    ##### Plot functions ######

    @add_fig_kwargs
    def plot_spectral_function(self,broadening=1,ax=None,with_S_nu=False,**kwargs):
        """
        Plot the Huang-Rhys spectral function S_hbarOmega

        Args:
            broadening: fwhm of the gaussian broadening in meV
            with_S_nu: True to add stem lines associated to the individuals partial Huang-Rhys factors
        """


        ax, fig, plt = get_ax_fig_plt(ax=ax)
        S_nu=self.S_nu()
        omega_nu=self.ph_eigfreq
        S_x,S_y=self.S_hbarOmega(broadening=broadening)

        ax.plot(S_x,S_y,**kwargs)
        ax.set_xlabel('Phonon energy (eV)')
        ax.set_ylabel(r'$S(\hbar\omega)$  (1/eV)')

        if with_S_nu==True:
            ax2=ax.twinx()
            markerline, stemlines, baseline = ax2.stem(omega_nu,S_nu,markerfmt='o',basefmt="k")
            plt.setp(markerline,'color',color)
            plt.setp(stemlines, 'color', plt.getp(markerline, 'color'))
            plt.setp(stemlines, 'linestyle', 'dotted')
            ax2.set_ylabel(r'$S_{\nu}$')

        return fig

    @add_fig_kwargs
    def plot_emission_spectrum(self,unit='eV',T=0,lamb=3,w=3,ax=None,**kwargs):
        """
        Plot the Luminescence intensity

        Args:
            unit: 'eV', 'cm-1', or 'nm'
            T: Temperature in K
            lamb: Lorentzian broadening applied to the vibronic peaks, in meV
            w: Gaussian broadening applied to the vibronic peaks, in meV
        """

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        x_eV,y_eV=self.L_hw(T=T,lamb=lamb,w=w) # in eV

        x_cm=x_eV*8065.73
        y_cm=y_eV/8065.73

        x_nm=1239.84193/x_eV
        y_nm=y_eV*(x_eV**2/1239.84193)


        if unit=='eV':
            ax.plot(x_eV,y_eV,**kwargs)
            ax.set_xlabel('Photon energy (eV)')
            ax.set_ylabel(r'$L(\hbar\omega)$  (1/eV)')


        elif unit=='cm-1':
            ax.plot(x_cm,y_cm,**kwargs)
            ax.set_xlabel(r'Photon energy ($cm^{-1}$)')
            ax.set_ylabel(r'$L(\hbar\omega)$  (1/$cm^{-1}$)')

        elif unit=='nm':
            ax.plot(x_nm,y_nm,**kwargs)
            ax.set_xlabel(r'Photon wavelength (nm))')
            ax.set_ylabel(r'Intensity (a.u.)')

        return fig


def get_forces_on_phonon_supercell(dSCF_supercell,phonon_supercell,forces_dSCF,tol):
    forces_in_supercell = np.zeros(shape=(len(phonon_supercell), 3))
    mapping=get_matching_dSCF_phonon_spcell(dSCF_supercell,phonon_supercell,tol)
    for i in range(len(mapping)):
        forces_in_supercell[mapping[i]]=forces_dSCF[i]

    return forces_in_supercell

def get_displacements_on_phonon_supercell(dSCF_supercell,phonon_supercell,displacements_dSCF,tol):
    displacements_in_supercell=np.zeros(shape=(len(phonon_supercell), 3))
    mapping=get_matching_dSCF_phonon_spcell(dSCF_supercell,phonon_supercell,tol)
    for i in range(len(mapping)):
        displacements_in_supercell[mapping[i]]=displacements_dSCF[i]

    return displacements_in_supercell

def get_matching_dSCF_phonon_spcell(dSCF_spcell,phonon_spcell,tol):
    dSCF_spcell_cart=dSCF_spcell.cart_coords
    phonon_spcell_cart=phonon_spcell.cart_coords
    # perform the matching
    mapping = []
    for i, site_1 in enumerate(dSCF_spcell):  # subset structure
        for j, site_2 in enumerate(phonon_spcell):  # superset structure
            if max(abs(dSCF_spcell_cart[i] - phonon_spcell_cart[j])) < tol :
                mapping.append(j)

    if len(mapping)==len(dSCF_spcell):
        print("Mapping between delta SCF supercell and phonon supercell succeeded. ",len(mapping),"/",len(dSCF_spcell))
    if len(mapping)!=len(dSCF_spcell):
        print("Caution... Mapping between delta SCF supercell and phonon supercell did not succeed. ",len(mapping),"/",len(dSCF_spcell))

    return mapping



#!/usr/bin/env python
"""Analyze the scalability of the OpenMP sections in the GS part. 1 k-point, cg method."""
import sys
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """Build a template input file for GS calculations with k-point parallelism """
    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc") if not paw else \
              abidata.pseudos("Si.GGA_PBE-JTH-paw.xml", "o.paw")
    structure = abidata.structure_from_ucell("SiO2-alpha")

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[1,1,1], shiftk=[0,0,0])

    # Global variables
    ecut = 24
    inp.set_vars(
        ecut=ecut,
        pawecutdg=ecut*2 if paw else None,
        nsppol=1,
        nband=28,
        paral_kgb=0,
        #istwfk="*1",
        #fftalg=312,
        timopt=-1,
        chksymbreak=0,
        prtwf=0,
        prtden=0,
        tolvrs=1e-10,
    )

    return inp


def build_flow(options):
    inp = make_input(paw=options.paw)
    nkpt = len(inp.abiget_ibz().points)

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)
    work = flowtk.Work()

    omp_list = options.omp_list
    if omp_list is None: omp_list = [1, 2, 4, 6]
    print("Using omp_list:", omp_list)

    mpi_procs = 1
    for omp_threads in omp_list:
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        work.register(inp, manager=manager)

    flow.register_work(work)
    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Pseudopotential Generators."""
from __future__ import annotations

#import abc
#import os
#import tempfile
#import collections
#import shutil
#import time

#from shutil import which
#from monty.termcolor import cprint
#from abipy.flowtk.pseudos import Pseudo
from abipy.ppcodes.base_parser import BaseParser

import logging
logger = logging.getLogger(__name__)


class AtompawParser(BaseParser):

    def scan(self, verbose: int = 0) -> AtompawParser:
        """
        Scan the output directory, set `run_completed` attribute.

        Raises: self.Error if invalid file.
        """
        try:
            self._scan(verbose=verbose)
        except Exception as exc:
            raise self.Error(f"Exception while parsing: {self.filepath}") from exc

    #def _select_by_prefix(self, prefix: str) -> list:
    #    return [p for p in self.all_filepaths if p.startswith(prefix)]

    def _scan(self, verbose: int=0) -> AtompawParser:
        if not os.path.exists(self.filepath):
            raise self.Error(f"File {self.filepath} does not exist")

        self.all_filepaths = os.listdir(self.workdir)

        # parse AE and PS wavefunctions.

        # aepsi.1





#!/usr/bin/env python
"""
Script to download and install pseudopotential tables from the web.
"""
from __future__ import annotations

import sys
#import os
import argparse
import abipy.tools.cli_parsers as cli

from monty.termcolor import cprint
from pymatgen.core.periodic_table import Element
from abipy.core.release import __version__
from abipy.tools import duck
from abipy.flowtk.pseudos import PseudoTable
from abipy.flowtk.psrepos import (tabulate_repos, repos_from_names,
                                  get_all_registered_repos, get_installed_repos_and_root)


def abips_list(options) -> list:
    """
    List installed pseudopotential repos.
    """
    repos, repos_root = get_installed_repos_and_root()
    if not repos:
        print("Could not find any pseudopotential repository installed in:", repos_root)
        return 0

    print(f"The following pseudopotential repositories are installed in {repos_root}:\n")
    print(tabulate_repos(repos, exclude=["installed"], verbose=options.verbose))

    if not options.checksums:
        print("\nUse -c to validate the md5 checksum")
        return 0

    exc_list = []
    for repo in repos:
        try:
            repo.validate_checksums(options.verbose)
        except Exception as exc:
            exc_list.append(exc)

    if exc_list:
        print("\nList of exceptions raised by validate_checksums:")
        for exc in exc_list:
            print(exc)

    return len(exc_list)


def abips_avail(options) -> int:
    """
    Show available pseudopotential repos.
    """
    print("List of available pseudopotential repositories:\n")
    all_repos = get_all_registered_repos()
    print(tabulate_repos(all_repos, with_citations=True, verbose=options.verbose))
    return 0


#def abips_nc_install(options) -> int:
#    """
#    Get all NC repos for a given version,
#    Can choose among three formats: psp8, upf2 and psml. By default we fetch all formats.
#    """
#    repos_root = get_repos_root(options)
#    all_repos = get_all_registered_repos()
#    repos = [repo for repo in all_repos if repo.isnc and not repo.is_installed(repos_root)]
#    if not repos:
#        print(f"All registered NC repositories are already installed in {repos_root}. Returning")
#        return 0
#
#    print("The following NC repositories will be installed:\n")
#    pprint_repos(repos, repos_root=repos_root)
#    if not options.yes and cli.user_wants_to_abort():
#        return 2
#
#    print("Fetching NC repositories. It may take some time ...")
#    for repo in repos:
#        repo.install(repos_root, options.verbose)
#
#    if options.verbose: abips_list(options)
#    return 0
#
#
#def abips_paw_install(options):
#    """
#    Get all JTH PAW repositories in PAWXML format for the given version.
#    """
#    repos_root = get_repos_root(options)
#    all_repos = get_all_registered_repos()
#    repos = [repo for repo in ALL_REPOS if repo.ispaw and not repo.is_installed(repos_root)]
#    if not repos:
#        print(f"All registered PAW repositories are already installed in {repos_root}. Returning")
#        return 0
#
#    print("The following PAW repositories will be installed:")
#    pprint_repos(repos, repos_root=repos_root)
#    if not options.yes and cli.user_wants_to_abort():
#        return 2
#
#    print("Fetching PAW repositories. It may take some time ...")
#    for repo in repos:
#        repo.install(repos_root, options.verbose)
#
#    if options.verbose: abips_list(options)
#    return 0


def abips_install(options) -> int:
    """
    Install pseudopotential repositories by name(s).
    Use `avail` command to get repo names.
    """
    repos = repos_from_names(options.repo_names)
    repos = [repo for repo in repos if not repo.is_installed()]

    if not repos:
        print("Table(s) are already installed! Nothing to do. Returning")
        return 0

    print("The following pseudopotential repositories will be installed:")
    print(tabulate_repos(repos, verbose=options.verbose), "\n")
    #if not options.yes and cli.user_wants_to_abort(): return 2

    for repo in repos:
        repo.install(verbose=options.verbose)

    if options.verbose:
        abips_list(options)

    return 0


def abips_show(options) -> int:
    """
    Show info on pseudopotential table(s).
    """
    repos = repos_from_names(options.repo_names)
    repos = [repo for repo in repos if repo.is_installed()]

    if not repos:
        print(f"There's no installed repository with name in: {options.repo_names}")
        return 1

    for repo in repos:
        print(repo)
        for table_name in repo.table_names:
            pseudos = repo.get_pseudos(table_name=table_name)
            print(f"For table_name: {table_name}:\n")
            if options.symbol is not None:
                print("Selecting pseudos with symbol:", options.symbol)
                pseudos = pseudos.pseudo_with_symbol(options.symbol, allow_multi=False)
            print(pseudos)

    return 0


def abips_element(options) -> int:
    """
    Find all pseudos in the installed tables for the given element (symbol or znucl).
    """
    # Accept symbol string or Z
    symbol = options.element
    if symbol.isnumeric():
        symbol = Element.from_Z(int(symbol)).symbol

    repos, repos_root = get_installed_repos_and_root()
    if not repos:
        print("Could not find any pseudopotential repository installed in:", repos_root)
        return 1

    pseudo_list = []
    for repo in repos:
        for table_name in repo.table_names:
            try:
                pseudos = repo.get_pseudos(table_name=table_name)
                pseudo = pseudos.pseudo_with_symbol(symbol, allow_multi=False)
            except Exception as exc:
                cprint(f"{str(exc)}", "red")
                continue

            if pseudo not in pseudo_list:
                pseudo_list.append(pseudo)

    pseudos = PseudoTable(pseudo_list)
    pseudos.print_table()

    return 0


def abips_mkff(options) -> int:
    """
    Call Abinit to compute PSPS.nc files from a list of pseudos and show results.
    """
    from abipy.electrons.psps import PspsFile, PspsRobot
    ecut = options.ecut

    if len(options.pseudo_paths) == 1:
        if options.vloc_rcut_list is None:
            with PspsFile.from_abinit_run(options.pseudo_paths[0], ecut) as abifile:
                if options.verbose: print(abifile)
                abifile.expose(use_web=options.expose_web,
                               slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                               verbose=options.verbose
                               )

        else:
            robot = PspsRobot.from_vloc_rcut_list(options.pseudo_paths[0], options.vloc_rcut_list, ecut)
            if options.verbose: print(robot)
            robot.expose(use_web=options.expose_web,
                         slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                         verbose=options.verbose
                         )

    else:
        if options.vloc_rcut_list is not None:
            raise ValueError("vloc_rcut_list does not support more than one pseudo!")
        robot = PspsRobot.from_abinit_run(options.pseudo_paths, ecut)

        if options.verbose: print(robot)
        robot.expose(use_web=options.expose_web,
                     slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                     verbose=options.verbose
                     )

    return 0


def get_epilog() -> str:
    return """\

Usage example:

  abips.py avail                             --> Show all registered pseudopotential repositories.
  abips.py list                              --> List repositories installed on this machine.
  abips.py install ONCVPSP-PBEsol-SR-PDv0.4  --> Install repository by name (requires internet connection).
  abips.py show ONCVPSP-PBEsol-SR-PDv0.4     --> Show info on pseudos in repository.
  abips.py element O                         --> Show all installed pseudos for element.
  abips.py mkff PSEUDO1 [PSEUDO2 ...]        --> Compute form factors for pseudos and show them.
"""
  #abips.py onc_install                        --> Get all NC repositories (most recent version)
  #abips.py onc_install -xc PBE -fr -sr -v 0.4
  #abips.py jth_install                        --> Get all JTH PAW repositories (most recent version)


def get_parser(with_epilog=False):

    # Parent parser for common options.
    copts_parser = argparse.ArgumentParser(add_help=False)
    copts_parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                              help='verbose, can be supplied multiple times to increase verbosity.')

    copts_parser.add_argument('--loglevel', default="ERROR", type=str,
                              help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    #copts_parser.add_argument('--repos-root', "-r", type=str,
    #                          default=os.path.expanduser(os.path.join("~", ".abinit", "pseudos")),
    #                          help='Installation directory. Default: $HOME/.abinit/pseudos')

    #copts_parser.add_argument('-y', "--yes", action="store_true", default=False,
    #                          help="Do not ask for confirmation when installing repositories.")

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version=__version__)

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    # Subparser for avail command.
    subparsers.add_parser("avail", parents=[copts_parser], help=abips_avail.__doc__)

    # Subparser for list command.
    p_list = subparsers.add_parser("list", parents=[copts_parser], help=abips_list.__doc__)
    p_list.add_argument("-c", "--checksums", action="store_true", default=False,
                        help="Validate md5 checksums.")

    # Subparser for install command.
    p_install = subparsers.add_parser("install", parents=[copts_parser], help=abips_install.__doc__)
    p_install.add_argument("repo_names", type=str, nargs="+", help="List of repositories to download.")
    p_install.add_argument("-c", "--checksums", action="store_true", default=False,
                           help="Validate md5 checksums.")

    # Subparser for onc_install command.
    #p_onc_install = subparsers.add_parser("onc_install", parents=[copts_parser], help=abips_onc_install.__doc__)
    #p_onc_install.add_argument("-v", type=str, default=None, help="Table version. Default: latest one ")

    # Subparser for jth_install command.
    #p_jth_install = subparsers.add_parser("jth_install", parents=[copts_parser], help=abips_paw_install.__doc__)
    #p_jth_install.add_argument("-v", type=str, default=None, help="Table version. Default: latest one ")

    # Subparser for show command.
    p_show = subparsers.add_parser("show", parents=[copts_parser], help=abips_show.__doc__)
    p_show.add_argument("repo_names", type=str, nargs="+", help="List of repo names.")
    p_show.add_argument("-s", "--symbol", type=str, default=None, help="Select pseudo by element symbol.")

    # Subparser for list command.
    p_element = subparsers.add_parser("element", parents=[copts_parser], help=abips_element.__doc__)
    p_element.add_argument("element", type=str, help="Element symbol or atomic number.")

    # Subparser for mkff command.
    p_mkff = subparsers.add_parser("mkff", parents=[copts_parser], help=abips_mkff.__doc__)
    p_mkff.add_argument("pseudo_paths", nargs="+", type=str, help="Pseudopotential path.")
    p_mkff.add_argument("--ecut", type=float, required=True, help="Cutoff energy in Ha.")
    p_mkff.add_argument("-rc", "--vloc-rcut-list", nargs="+", default=None, type=float,
                        help="List of cutoff radii for vloc in Bohr.")
    cli.add_expose_options_to_parser(p_mkff)

    return parser


def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        print(exc)
        show_examples_and_exit(error_code=1)

    #cli.set_loglevel(options.loglevel)

    # Use seaborn settings.
    if hasattr(options, "seaborn") and options.seaborn:
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)

    return globals()[f"abips_{options.command}"](options)


if __name__ == "__main__":
    sys.exit(main())



"""
Interface to the GKQ.nc file storing the e-ph matrix elements
in the atomic representation (idir, ipert) for a single q-point.
This file is produced by the eph code with eph_task -4.

To analyze the e-ph scattering potentials, use v1qavg and eph_task 15 or -15
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from collections import OrderedDict
from monty.string import marquee
from monty.functools import lazy_property
from abipy.core.structure import Structure
from abipy.core.kpoints import Kpoint
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt
from abipy.tools import duck
from abipy.tools.typing import Figure, PathLike
from abipy.abio.robots import Robot
from abipy.electrons.ebands import ElectronBands, ElectronsReader, RobotWithEbands
from abipy.eph.common import glr_frohlich, EPH_WTOL


class GkqFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):

    @classmethod
    def from_file(cls, filepath: PathLike):
        """Initialize the object from a netcdf_ file."""
        return cls(filepath)

    def __init__(self, filepath: PathLike):
        super().__init__(filepath)
        self.r = GkqReader(filepath)

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))
        app("qpoint in g(k,q): %s" % str(self.qpoint))
        app("uses_interpolated_dvdb: %s" % str(self.uses_interpolated_dvdb))
        app("phonon frequencies in Ha %s:" % str(self.phfreqs_ha))
        app("Macroscopic dielectric tensor in Cartesian coordinates:")
        app(str(self.epsinf_cart))
        app("")
        app("Born effective charges in Cartesian coordinates:")
        for i, (site, bec) in enumerate(zip(self.structure, self.becs_cart)):
            app("[%d]: %s" % (i, repr(site)))
            app(str(bec))
            app("")

        app(r"Fulfillment of charge neutrality, F_{ij} = \sum_{atom} Z^*_{ij,atom}:")
        f = np.sum(self.becs_cart, axis=0)
        app(str(f) + "\n")

        return "\n".join(lines)

    def close(self) -> None:
        self.r.close()

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.r.read_ebands()

    @lazy_property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def uses_interpolated_dvdb(self) -> bool:
        """True if the matrix elements have been computed with an interpolated potential."""
        return int(self.r.read_value("interpolated")) == 1

    @lazy_property
    def params(self) -> dict:
        """Dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    @lazy_property
    def qpoint(self) -> Kpoint:
        """Q-point object."""
        return Kpoint(self.r.read_value('qpoint'), self.structure.reciprocal_lattice)

    @lazy_property
    def phfreqs_ha(self) -> np.ndarray:
        """(3 * natom) array with phonon frequencies in Ha."""
        return self.r.read_value("phfreqs")

    @lazy_property
    def phdispl_cart_bohr(self) -> np.ndarray:
        """(natom3_nu, natom3) complex array with the phonon displacement in cartesian coordinates in Bohr."""
        return self.r.read_value("phdispl_cart", cmode="c")

    @lazy_property
    def phdispl_red(self) -> np.ndarray:
        """(natom3_nu, natom3) complex array with the phonon displacement in reduced coordinates."""
        return self.r.read_value("phdispl_red", cmode="c")

    @lazy_property
    def becs_cart(self) -> np.ndarray:
        """(natom, 3, 3) array with the Born effective charges in Cartesian coordinates."""
        return self.r.read_value("becs_cart").transpose(0, 2, 1).copy()

    @lazy_property
    def epsinf_cart(self) -> np.ndarray:
        """(3, 3) array with electronic macroscopic dielectric tensor in Cartesian coordinates."""
        return self.r.read_value("emacro_cart").T.copy()

    @lazy_property
    def eigens_kq(self) -> np.ndarray:
        """(spin, nkpt, mband) array with eigenvalues on the k+q grid in eV."""
        return self.r.read_value("eigenvalues_kq") * abu.Ha_eV

    @staticmethod
    def _check_mode(mode: str) -> None:
        """Check whether mode is allowed."""
        allowed_modes = ("atom", "phonon")
        if mode not in allowed_modes:
            raise ValueError(f"Invalid {mode=}, it should be in {allowed_modes=}")

    def read_gkq_kpoint(self, kpoint, mode: str = "phonon") -> np.ndarray:
        """
        Read e-ph matrix stored on disk for all spins and the given k-point

        Args:
            kpoint:
            mode: "phonon" for e-ph matrix elements in phonon representation,
                  "atom" for e-ph matrix elements in the atomic representation (idir, iatom).

        Return: complex array with shape: (nsppol, 3*natom, mband, mband)
                                                            m_kq, n_k    <-- band indices.
        """
        self._check_mode(mode)

    def read_all_gkq(self, mode: str = "phonon") -> np.ndarray:
        """
        Read all e-ph matrix stored on disk.

        Args:
            mode: "phonon" for e-ph matrix elements in phonon representation,
                  "atom" for e-ph matrix elements in the atomic representation (idir, iatom).

        Return: complex array with shape: (nsppol, nkpt, 3*natom, mband, mband)
                                                                   m_kq, n_k    <-- band indices.
        """
        self._check_mode(mode)

        # Read the e-ph matrix element in the atomic representation (idir, ipert). Fortran array on disk has shape:
        # nctkarr_t('gkq', "dp",
        #           'complex, max_number_of_states, max_number_of_states, number_of_phonon_modes, number_of_kpoints, number_of_spins')
        # The first band index in Fortran refers to m_kq, the second one to n_k.
        # hence we have to transpose the (nb_kq, nb_k) submatrix written by ABINIT.
        gkq_atm = self.r.read_value("gkq", cmode="c").transpose(0, 1, 2, 4, 3).copy()
        if mode == "atom":
            return gkq_atm

        # Convert from atomic to phonon representation. May use np.einsum for better efficiency but oh well!
        nband = gkq_atm.shape[-1]
        assert nband == gkq_atm.shape[-2] and nband == self.ebands.nband
        nb2, natom3 = nband ** 2, 3 * len(self.structure)
        phfreqs_ha, phdispl_red = self.phfreqs_ha, self.phdispl_red

        gkq_nu, cwork = np.empty_like(gkq_atm), np.empty((natom3, nb2), dtype=complex)
        for spin in range(self.ebands.nsppol):
            for ik in range(self.ebands.nkpt):
                gc = np.reshape(gkq_atm[spin, ik], (-1, nb2))
                for nu in range(natom3):
                    cwork[nu] = np.dot(phdispl_red[nu], gc) / np.sqrt(2.0 * phfreqs_ha[nu]) if (phfreqs_ha[nu] > EPH_WTOL) else 0.0
                gkq_nu[spin, ik] = np.reshape(cwork, (natom3, nband, nband))

        return gkq_nu

    def get_absg_kpoint(self, kpoint, eps_mev: float=0.01) -> tuple[np.ndarray, np.ndarray, int, Kpoint]:
        """
        Args:
            kpoint: |Kpoint| object or list/tuple with reduced coordinates or integer with the index
            eps_mev: Tolerance in mev used to detect degeneracies
        """
        if duck.is_intlike(kpoint):
            ik = kpoint
            kpoint = self.kpoints[ik]
        else:
            kpoint = Kpoint.as_kpoint(kpoint, self.structure.reciprocal_lattice)
            ik = self.kpoints.index(kpoint)

        eps_ha = eps_mev / abu.Ha_meV
        eps_ev = eps_ha * abu.Ha_eV

        nsppol = self.ebands.nsppol
        natom3 = len(self.structure) * 3
        nb = self.ebands.nband

        phfreqs_ha = self.phfreqs_ha
        eigens_k = self.ebands.eigens
        eigens_kq = self.eigens_kq

        # (nsppol, nkpt, 3*natom, mband, mband) real array.
        absg = np.abs(self.read_all_gkq(mode="phonon")) * abu.Ha_meV
        absgk = absg[:,ik].copy()
        absg_unsym = absg[:,ik].copy()
        absg_sym = np.zeros_like(absgk)

        # Average over phonons.
        for spin in range(nsppol):
            g2_mn = np.zeros((nb, nb), dtype=float)
            for nu in range(natom3):
                w_1 = phfreqs_ha[nu]
                g2_mn[:], nn = 0.0, 0
                for mu in range(natom3):
                    w_2 = phfreqs_ha[mu]
                    if abs(w_1 - w_2) >= eps_ha: continue
                    nn += 1
                    g2_mn += absgk[spin,mu,:,:] ** 2
                absg_sym[spin,nu,:,:] = np.sqrt(g2_mn / nn)

        # Average over k electrons.
        absg = absg_sym.copy()
        g2_nu = np.zeros((natom3), dtype=float)
        for spin in range(nsppol):
            for jbnd in range(nb):
                for ibnd in range(nb):
                    w_1 = eigens_k[spin, ik, ibnd]
                    g2_nu[:], nn = 0.0, 0
                    for pbnd in range(nb):
                        w_2 = eigens_k[spin, ik, pbnd]
                        if abs(w_2 - w_1) >= eps_ev: continue
                        nn += 1
                        # MG FIXME: Why absgk and not absg here as done below for k+q?
                        g2_nu += absgk[spin,:,jbnd,pbnd] ** 2
                    absg_sym[spin,:,jbnd,ibnd] = np.sqrt(g2_nu / nn)

        # Average over k+q electrons.
        absgk = absg_sym.copy()
        for spin in range(nsppol):
            for ibnd in range(nb):
                for jbnd in range(nb):
                    w_1 = eigens_kq[spin, ik, jbnd]
                    g2_nu[:], nn = 0.0, 0
                    for pbnd in range(nb):
                        w_2 = eigens_kq[spin, ik, pbnd]
                        if abs(w_2 - w_1) >= eps_ev: continue
                        nn += 1
                        g2_nu += absgk[spin,:,pbnd,ibnd] ** 2
                    absg_sym[spin,:,jbnd,ibnd] = np.sqrt(g2_nu / nn)

        return absg_sym, absg_unsym, ik, kpoint

    def get_qe_dataframe(self, kpoint) -> pd.DataFrame:
        """
        Build and return a dataframe with |g(k,q)|^2 for the given k-point and all bands.

        Args:
            kpoint:
        """
        absg, absg_unsym, ik, kpoint = self.get_absg_kpoint(kpoint)

        # Now insert absg array in a pandas dataframe.
        # Flatten the array, get the indices and combine indices and values into a DataFrame
        shape, ndim = absg.shape, absg.ndim
        indices = np.indices(shape).reshape(ndim, -1).T
        df = pd.DataFrame(indices, columns=["spin", "imode", "m_kq", "n_k"])
        df["|g|[meV]"] = absg.flatten()
        df["ik"] = ik

        # Add columns with phonon frequencies and electron energies in meV at k and k+q.
        imodes = df["imode"].to_numpy()
        df["omega(q)[meV]"] = (self.phfreqs_ha * abu.Ha_meV)[imodes]
        spin_inds, mkq_inds, nk_inds = df["spin"].to_numpy(), df["m_kq"].to_numpy(), df["n_k"].to_numpy()
        #print(self.ebands.eigens[spin_inds,ik,nk_inds].shape)
        df["e_nk[eV]"] = self.ebands.eigens[spin_inds, ik, nk_inds]
        df["e_mkq[eV]"] = self.eigens_kq[spin_inds, ik, mkq_inds]

        # Reorder the columns and drop the index
        new_order = ["n_k", "m_kq", "spin", "imode", "e_nk[eV]", "e_mkq[eV]", "omega(q)[meV]", "|g|[meV]"]
        return df[new_order].reset_index(drop=True)

    @add_fig_kwargs
    def plot(self, mode="phonon", with_glr=True, fontsize=8, colormap="viridis", sharey=True, **kwargs) -> Figure:
        """
        Plot the gkq matrix elements for a given q-point.

        Args:
            mode: "phonon" to plot eph matrix elements in the phonon representation,
                  "atom" for atomic representation.
            with_glr: True to plot the long-range component estimated from Verdi's model.
            fontsize: Label and title fontsize.
            colormap: matplotlib colormap
            sharey: True if yaxis should be shared among axes.

        Return: |matplotlib-Figure|
        """
        gkq = np.abs(self.read_all_gkq(mode=mode))
        if mode == "phonon": gkq *= abu.Ha_meV

        # Compute e_{k+q} - e_k for all possible (b, b')
        ediffs = np.empty_like(gkq)
        for spin in range(self.ebands.nsppol):
            for ik in range(self.ebands.nkpt):
                for ib_kq in range(self.ebands.mband):
                    for ib_k in range(self.ebands.mband):
                        ed = np.abs(self.eigens_kq[spin, ik, ib_kq] - self.ebands.eigens[spin, ik, ib_k])
                        ediffs[spin, ik, :, ib_k, ib_kq] = ed

        if with_glr and mode == "phonon":
            # Add horizontal bar with matrix elements computed from Verdi's model (only G = 0, \delta_nm in bands).
            dcart_bohr = self.phdispl_cart_bohr
            #dcart_bohr = self.r.read_value("phdispl_cart_qvers", cmode="c").real
            gkq_lr = glr_frohlich(self.qpoint, self.becs_cart, self.epsinf_cart,
                                  dcart_bohr, self.phfreqs_ha, self.structure)
            # self.phdispl_cart_bohr, self.phfreqs_ha, self.structure)
            gkq2_lr = np.abs(gkq_lr) * abu.Ha_meV

        natom = len(self.structure)
        num_plots, ncols, nrows = 3 * natom, 3, natom
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=sharey, squeeze=False)
        ax_list = ax_list.ravel()
        cmap = plt.get_cmap(colormap)

        for nu, ax in enumerate(ax_list):
            idir = nu % 3
            iat = (nu - idir) // 3
            data, c = gkq[:, :, nu, :, :].ravel(), ediffs[:,:,nu,:,:].ravel()
            # Filter items according to ediff
            index = c <= 1.2 * self.phfreqs_ha.max() * abu.Ha_eV
            data, c = data[index], c[index]
            sc = ax.scatter(np.arange(len(data)), data, alpha=0.9, s=30, c=c, cmap=cmap)
                            #facecolors='none', edgecolors='orange')
            plt.colorbar(sc, ax=ax)

            ax.grid(True)
            if iat == natom - 1:
                ax.set_xlabel("Matrix element index")
            if idir == 0:
                ylabel = r"$|g^{atm}_{\bf q}|$" if mode == "atom" else r"$|g_{\bf q}|$ (meV)"
                ax.set_ylabel(ylabel)

            ax.set_title(r"$\nu$: %d, $\omega_{{\bf q}\nu}$ = %.2E (meV)" %
                         (nu, self.phfreqs_ha[nu] * abu.Ha_meV), fontsize=fontsize)

            if with_glr:
                ax.axhline(gkq2_lr[nu], color='k', linestyle='dashed', linewidth=2)

        fig.suptitle("qpoint: %s" % repr(self.qpoint), fontsize=fontsize)
        return fig

    @add_fig_kwargs
    def plot_diff_with_other(self, other, mode="phonon", ax_list=None, labels=None, fontsize=8, **kwargs) -> Figure:
        """
        Produce scatter plot and histogram to compare the gkq matrix elements stored in two files.

            other: other GkqFile instance.
            mode: "phonon" to plot eph matrix elements in the phonon representation,
                  "atom" for atomic representation.
            ax_list: List with 2 matplotlib axis. None if new ax_list should be created.
            labels: Labels associated to self and other
            fontsize: Label and title fontsize.

        Return: |matplotlib-Figure|
        """
        if self.qpoint != other.qpoint:
            raise ValueError("Found different q-points: %s and %s" % (self.qpoint, other.qpoint))

        if labels is None:
            labels = ["this (interpolated: %s)" % self.uses_interpolated_dvdb,
                      "other (interpolated: %s)" % other.uses_interpolated_dvdb]

        this_gkq = np.abs(self.read_all_gkq(mode=mode))
        other_gkq = np.abs(other.read_all_gkq(mode=mode))
        if mode == "phonon":
            this_gkq *= abu.Ha_meV
            other_gkq *= abu.Ha_meV

        absdiff_gkq = np.abs(this_gkq - other_gkq)

        stats = OrderedDict([
            ("min", absdiff_gkq.min()),
            ("max", absdiff_gkq.max()),
            ("mean", absdiff_gkq.mean()),
            ("std", absdiff_gkq.std()),
        ])

        num_plots, ncols, nrows = 2, 2, 1
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # Downsample datasets. Show only points with error > threshold.
        ntot = absdiff_gkq.size
        threshold = stats["mean"] + stats["std"]
        data = this_gkq[absdiff_gkq > threshold].ravel()
        nshown = len(data)
        xs = np.arange(len(data))

        ax = ax_list[0]
        ax.scatter(xs, data, alpha=0.9, s=30, label=labels[0], facecolors='none', edgecolors='orange')

        data = other_gkq[absdiff_gkq > threshold].ravel()
        ax.scatter(xs, data, alpha=0.3, s=10, marker="x", label=labels[1], facecolors="g", edgecolors="none")

        ax.grid(True)
        ax.set_xlabel("Matrix element index")
        ylabel = r"$|g^{atm}_{\bf q}|$" if mode == "atom" else r"$|g_{\bf q}|$ (meV)"
        ax.set_ylabel(ylabel)
        ax.set_title(r"qpt: %s, $\Delta$ > %.1E (%.1f %%)" % (
                     repr(self.qpoint), threshold, 100 * nshown / ntot), fontsize=fontsize)
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        ax = ax_list[1]
        ax.hist(absdiff_gkq.ravel(), facecolor='g', alpha=0.75)
        ax.grid(True)
        ax.set_xlabel("Absolute Error" if mode == "atom" else "Absolute Error (meV)")
        ax.set_ylabel("Count")

        ax.axvline(stats["mean"], color='k', linestyle='dashed', linewidth=1)
        _, max_ = ax.get_ylim()
        ax.text(0.7, 0.7,  "\n".join("%s = %.1E" % item for item in stats.items()),
                fontsize=fontsize, horizontalalignment='center', verticalalignment='center',
                transform=ax.transAxes)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        yield self.plot()

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("gkq = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(gkq)"),
            nbv.new_code_cell("gkq.ebands.plot();"),
            nbv.new_code_cell("gkq.epsinf_cart;"),
            nbv.new_code_cell("gkq.becs_cart;"),
            nbv.new_code_cell("""
              #with abilab.abiopen('other_GKQ.nc') as other:
              #     gkq.plot_diff_with_other(other);
            """)
        ])

        return self._write_nb_nbpath(nb, nbpath)


class GkqReader(ElectronsReader):
    """
    This object reads the results stored in the GKQ file produced by ABINIT.
    It provides helper function to access the most important quantities.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GkqReader
    """


class GkqRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple GKQ.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GkqRobot
    """
    EXT = "GKQ"

    @lazy_property
    def kpoints(self):
        # Consistency check: kmesh should be the same in each file.
        ref_kpoints = self.abifiles[0].ebands.kpoints
        for i, abifile in enumerate(self.abifiles):
            if i == 0: continue
            if abifile.kpoints != ref_kpoints:
                for k1, k2 in zip(ref_kpoints, abifile.kpoints):
                    print("k1:", k1, "--- k2:", k2)
                raise ValueError("Found different list of kpoints in %s" % str(abifile.filepath))
        return ref_kpoints

    def _check_qpoints_equal(self):
        """Raises ValueError if different `qpoint` in files."""
        ref_qpoint = self.abifiles[0].qpoint
        for i, abifile in enumerate(self.abifiles):
            if i == 0: continue
            if abifile.qpoint != ref_qpoint:
                raise ValueError("Found different qpoint in %s" % str(abifile.filepath))

    #@add_fig_kwargs
    #def plot_gkq2_qpath(self, band_kq, band_k, kpoint=0, with_glr=False, qdamp=None, nu_list=None, # spherical_average=False,
    #                    ax=None, fontsize=8, eph_wtol=EPH_WTOL, **kwargs):
    #    ncols, nrows = 2, len(self) - 1
    #    num_plots = ncols * nrows
    #    ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,

    @add_fig_kwargs
    def plot_gkq2_qpath(self, band_kq, band_k, kpoint=0, with_glr=False, qdamp=None, nu_list=None, # spherical_average=False,
                        ax=None, fontsize=8, eph_wtol=EPH_WTOL, kq_labels=False, **kwargs) -> Figure:
        r"""
        Plot the magnitude of the electron-phonon matrix elements <k+q, band_kq| Delta_{q\nu} V |k, band_k>
        for a given set of (band_kq, band, k) as a function of the q-point.

        Args:
            band_ks: Band index of the k+q states (starts at 0)
            band_k: Band index of the k state (starts at 0)
            kpoint: |Kpoint| object or index.
            with_glr: True to plot the long-range component estimated from Verdi's model.
            qdamp:
            nu_list: List of phonons modes to be selected (starts at 0). None to select all modes.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: Label and title fontsize.
            kq_labels: If True, use the label associated to k+q instead of q.

        Return: |matplotlib-Figure|
        """
        if duck.is_intlike(kpoint):
            ik = kpoint
            kpoint = self.kpoints[ik]
        else:
            kpoint = Kpoint.as_kpoint(kpoint, self.abifiles[0].structure.reciprocal_lattice)
            ik = self.kpoints.index(kpoint)

        # Assume abifiles in the robot are already ordered according to q-path.
        xs = list(range(len(self.abifiles)))
        natom3 = len(self.abifiles[0].structure) * 3
        nsppol = self.abifiles[0].nsppol
        nqpt = len(self.abifiles)
        gkq_snuq = np.empty((nsppol, natom3, nqpt), dtype=complex)
        if with_glr: gkq_lr = np.empty((nsppol, natom3, nqpt), dtype=complex)

        # TODO: Should take into account possible degeneracies in k and k+q and phonon modes.
        xticks, xlabels = [], []
        for iq, abifile in enumerate(self.abifiles):
            qpoint = abifile.qpoint
            #d3q_fact = one if not spherical_average else np.sqrt(4 * np.pi) * qpoint.norm

            if kq_labels:
                name = abifile.structure.findname_in_hsym_stars(kpoint + qpoint)
            else:
                name = qpoint.name if qpoint.name is not None else abifile.structure.findname_in_hsym_stars(qpoint)

            if name is not None:
                xticks.append(iq)
                xlabels.append(name)

            phfreqs_ha, phdispl_red = abifile.phfreqs_ha, abifile.phdispl_red
            ncvar = abifile.r.read_variable("gkq")
            for spin in range(nsppol):
                gkq_atm = ncvar[spin, ik, :, band_k, band_kq]
                gkq_atm = gkq_atm[:, 0] + 1j * gkq_atm[:, 1]

                # Transform the gkq matrix elements from (atom, red_direction) basis to phonon-mode basis.
                gkq_snuq[spin, :, iq] = 0.0
                for nu in range(natom3):
                    if phfreqs_ha[nu] < eph_wtol: continue
                    gkq_snuq[spin, nu, iq] = np.dot(phdispl_red[nu], gkq_atm) / np.sqrt(2.0 * phfreqs_ha[nu])

            if with_glr:
                # Compute long range part with (simplified) generalized Frohlich model.
                gkq_lr[spin, :, iq] = glr_frohlich(qpoint, abifile.becs_cart, abifile.epsinf_cart,
                                                   abifile.phdispl_cart_bohr, phfreqs_ha, abifile.structure, qdamp=qdamp)

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        nu_list = list(range(natom3)) if nu_list is None else list(nu_list)
        for spin in range(nsppol):
            for nu in nu_list:
                ys = np.abs(gkq_snuq[spin, nu]) * abu.Ha_meV
                pre_label = kwargs.pop("pre_label",r"$g_{\bf q}$")
                if nsppol == 1: label = r"%s $\nu$: %s" % (pre_label, nu)
                if nsppol == 2: label = r"%s $\nu$: %s, spin: %s" % (pre_label, nu, spin)
                ax.plot(xs, ys, linestyle="--", label=label)
                if with_glr:
                    # Plot model with G = 0 and delta_nn'
                    ys = np.abs(gkq_lr[spin, nu]) * abu.Ha_meV
                    label = r"$g_{\bf q}^{\mathrm{lr0}}$ $\nu$: %s" % nu
                    ax.plot(xs, ys, linestyle="", marker="o", label=label)

        ax.grid(True)
        ax.set_xlabel("Wave Vector")
        ax.set_ylabel(r"$|g_{\bf q}|$ (meV)")
        if xticks:
            ax.set_xticks(xticks, minor=False)
            ax.set_xticklabels(xlabels, fontdict=None, minor=False, size=kwargs.pop("klabel_size", "large"))

        ax.legend(loc="best", fontsize=fontsize, shadow=True)
        title = r"$band_{{\bf k} + {\bf q}: %s, band_{\bf{k}}: %s, kpoint: %s" % (band_kq, band_k, repr(kpoint))
        ax.set_title(title, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_gkq2_diff(self, iref=0, **kwargs) -> Figure:
        """
        Wraps gkq.plot_diff_with_other
        Produce scatter and histogram plot to compare the gkq matrix elements stored in all the files
        contained in the robot. Assume all files have the same q-point. Compare the `iref` file with others.
        kwargs are passed to `plot_diff_with_other`.
        """
        if len(self) <= 1: return None
        self._check_qpoints_equal()

        ncols, nrows = 2, len(self) - 1
        num_plots = ncols * nrows
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=False, sharey=False, squeeze=False)

        ref_gkq, ref_label = self.abifiles[iref], self.labels[iref]
        cnt = -1
        for ifile, (other_label, other_gkq) in enumerate(zip(self.labels, self.abifiles)):
            if ifile == iref: continue
            cnt += 1
            labels = [ref_label, other_label]
            ref_gkq.plot_diff_with_other(other_gkq, ax_list=ax_mat[cnt], labels=labels, show=False, **kwargs)

        return fig

    def yield_figs(self, **kwargs): # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        for fig in self.get_ebands_plotter().yield_figs(): yield fig

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to `nbpath`. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.GkqRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("# robot.plot_gkq2_diff();"),
            nbv.new_code_cell("# robot.plot_gkq2_qpath(band_kq=0, band_k=0, kpoint=0, with_glr=True, qdamp=None);")
        ])

        # Mixins
        nb.cells.extend(self.get_baserobot_code_cells())
        nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)



#!/usr/bin/env python
"""
This benchmark compares the effective parallel efficiency with the one reported by autoparal.
"""
import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    """
    This function constructs the input files for the phonon calculation:
    GS input + the input files for the phonon calculation.
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")
    pseudos = abidata.pseudos("13al.981214.fhi", "33as.pspnc")
    if paw:
        raise NotImplementedError("PAW")

    # List of q-points for the phonon calculation.
    qpoints = [
             0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
             5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  2.50000000E-01,  0.00000000E+00,
            -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
            -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
            ]
    import numpy as np
    qpoints = np.reshape(qpoints, (-1,3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(
        nband=12,
        ecut=12.0,
        pawecutdg=24 if paw else None,
        ngkpt=[8, 8, 8],
        nshiftk=4,
        shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5],
        #shiftk=[0, 0, 0],
        paral_kgb=0,
        diemac=9.0,
    )

    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)
    gs_inp.set_vars(global_vars, tolvrs=1.0e-18)

    # Get the qpoints in the IBZ. Note that here we use a q-mesh with ngkpt=(4,4,4) and shiftk=(0,0,0)
    # i.e. the same parameters used for the k-mesh in gs_inp.
    #qpoints = gs_inp.abiget_ibz(ngkpt=(4,4,4), shiftk=(0,0,0), kptopt=1).points
    #print("get_ibz", qpoints)

    ph_inp = abilab.AbinitInput(structure, pseudos)

    # Response-function calculation for phonons.
    qpt = [0, 0, 0]
    qpt = [0.25, 0, 0]
    ph_inp.set_vars(
        global_vars,
        rfphon=1,        # Will consider phonon-type perturbation
        nqpt=1,          # One wavevector is to be considered
        qpt=qpt,         # This wavevector is q=0 (Gamma)
        toldfe=1.0e10,
        kptopt=3,
        timopt=-1,
    )
        #rfatpol   1 1   # Only the first atom is displaced
        #rfdir   1 0 0   # Along the first reduced coordinate axis
        #kptopt   2      # Automatic generation of k points, taking

    return gs_inp, ph_inp


def build_flow(options):
    gs_inp, ph_inp = make_inputs()

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)
    gs_work = flowtk.Work()
    gs_work.register_scf_task(gs_inp)
    flow.register_work(gs_work)
    flow.exclude_from_benchmark(gs_work)

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    print("Getting all autoparal confs up to max_ncpus:", max_ncpus, "with efficiency >=", min_eff)

    pconfs = ph_inp.abiget_autoparal_pconfs(max_ncpus, autoparal=1)
    if options.verbose: print(pconfs)

    omp_threads = 1
    work = flowtk.Work()
    for conf in pconfs:
        mpi_procs = conf.mpi_ncpus
        if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
        if min_eff is not None and conf.efficiency < min_eff: continue

        if options.verbose: print(conf)
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        inp = ph_inp.new_with_vars(conf.vars)
        work.register_phonon_task(inp, manager=manager, deps={gs_work[0]: "WFK"})

    print("Found %d configurations" % len(work))
    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
Benckmark for Molecular dynamics (based on tutoparal/Input/tmoldyn[1-3].in)
"""
import sys
import numpy as np
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """
    Build and return an input file for MD calculations with paral_kgb=1
    """
    pseudos = abidata.pseudos("Al.oncvpsp") if not paw else \
              abidata.pseudos("Al.GGA_PBE-JTH-paw.xml")

    # Atomic Positions.
    xred = np.fromstring("""
0.00000000E+00  0.00000000E+00  0.00000000E+00
0.00000000E+00  0.25000000E+00  0.25000000E+00
0.25000000E+00  0.00000000E+00  0.25000000E+00
0.25000000E+00  0.25000000E+00  0.00000000E+00
0.00000000E+00  0.00000000E+00  0.50000000E+00
0.00000000E+00  0.25000000E+00  0.75000000E+00
0.25000000E+00  0.00000000E+00  0.75000000E+00
0.25000000E+00  0.25000000E+00  0.50000000E+00
0.00000000E+00  0.50000000E+00  0.00000000E+00
0.00000000E+00  0.75000000E+00  0.25000000E+00
0.25000000E+00  0.50000000E+00  0.25000000E+00
0.25000000E+00  0.75000000E+00  0.00000000E+00
0.00000000E+00  0.50000000E+00  0.50000000E+00
0.00000000E+00  0.75000000E+00  0.75000000E+00
0.25000000E+00  0.50000000E+00  0.75000000E+00
0.25000000E+00  0.75000000E+00  0.50000000E+00
0.50000000E+00  0.00000000E+00  0.00000000E+00
0.50000000E+00  0.25000000E+00  0.25000000E+00
0.75000000E+00  0.00000000E+00  0.25000000E+00
0.75000000E+00  0.25000000E+00  0.00000000E+00
0.50000000E+00  0.00000000E+00  0.50000000E+00
0.50000000E+00  0.25000000E+00  0.75000000E+00
0.75000000E+00  0.00000000E+00  0.75000000E+00
0.75000000E+00  0.25000000E+00  0.50000000E+00
0.50000000E+00  0.50000000E+00  0.00000000E+00
0.50000000E+00  0.75000000E+00  0.25000000E+00
0.75000000E+00  0.50000000E+00  0.25000000E+00
0.75000000E+00  0.75000000E+00  0.00000000E+00
0.50000000E+00  0.50000000E+00  0.50000000E+00
0.50000000E+00  0.75000000E+00  0.75000000E+00
0.75000000E+00  0.50000000E+00  0.75000000E+00
0.75000000E+00  0.75000000E+00  0.50000000E+00
""", sep=" ").reshape((-1,3))

    # Crystal structure (32 Al atoms)
    structure = abilab.Structure.from_abivars(
        acell=3*[12.81],
        rprim=np.eye(3),
        ntypat=1,
        typat=32*[1],
        znucl=13.0,
        xred=xred,
    )

    inp = abilab.AbinitInput(structure, pseudos)

    inp.set_vars(
        # parallelization
        paral_kgb=1,
        #npband=10,
        #npfft=3,
        #npkpt=4,
        #bandpp=2,

        ecut=3.0,
        pawecutdg=6.0 if paw else None,
        nband=80,
        nsppol=1,

        # SCF cycle parameters
        tolvrs=1.e-3,
        nstep=50,

        # K-points and sym
        occopt=3,
        nsym=1,

        # Molecular Dynamics parameters
        ionmov=12,
        ntime=100,
        dtion=100,
        mdtemp=[3000, 3000],
        tsmear=0.009500446,

        # IO
        prtden=0,
        prtwf=0,
        prteig=0,
        timopt=-1,
    )

    inp.set_kmesh(ngkpt=[1,1,1], shiftk=[0,0,0])

#%% nprocs_to_test = 120

    return inp


def build_flow(options):
    template = make_input()
    #template.abivalidate()

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    if max_ncpus is None:
        nkpt = len(template.abiget_ibz().points)
        max_ncpus = nkpt * template["nsppol"] * template["nband"] * 4
    print("Getting all autoparal confs up to max_ncpus:", max_ncpus, "with efficiency >=", min_eff)

    pconfs = template.abiget_autoparal_pconfs(max_ncpus, autoparal=1, verbose=options.verbose)
    if options.verbose: print(pconfs)

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    work = flowtk.Work()
    for conf, omp_threads in product(pconfs, options.omp_list):
        mpi_procs = conf.mpi_ncpus
        if not options.accept_conf(conf, omp_threads): continue

        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        inp = template.new_with_vars(conf.vars)
        work.register_scf_task(inp, manager=manager)

    print("Found %d configurations" % len(work))
    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""Benchmark for Optic calculations."""
import sys
import os
import abipy.data as data
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_base_flow(options):
    multi = abilab.MultiDataset(structure=data.structure_from_ucell("GaAs"),
                                pseudos=data.pseudos("31ga.pspnc", "33as.pspnc"), ndtset=5)

    # Global variables
    kmesh = dict(ngkpt=[4, 4, 4],
                 nshiftk=4,
                 shiftk=[[0.5, 0.5, 0.5],
                         [0.5, 0.0, 0.0],
                         [0.0, 0.5, 0.0],
                         [0.0, 0.0, 0.5]]
    )

    paral_kgb = 1
    global_vars = dict(ecut=2, paral_kgb=paral_kgb)
    global_vars.update(kmesh)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_vars(
        tolvrs=1e-6,
        nband=4,
    )

    # NSCF run with large number of bands, and points in the the full BZ
    multi[1].set_vars(
        iscf=-2,
        nband=20,
        nstep=25,
        kptopt=1,
        tolwfr=1.e-9,
        #kptopt=3,
    )

    # Fourth dataset: ddk response function along axis 1
    # Fifth dataset: ddk response function along axis 2
    # Sixth dataset: ddk response function along axis 3
    for dir in range(3):
        rfdir = 3 * [0]
        rfdir[dir] = 1

        multi[2+dir].set_vars(
            iscf=-3,
            nband=20,
            nstep=1,
            nline=0,
            prtwf=3,
            kptopt=3,
            nqpt=1,
            qpt=[0.0, 0.0, 0.0],
            rfdir=rfdir,
            rfelfd=2,
            tolwfr=1.e-9,
        )

    scf_inp, nscf_inp, ddk1, ddk2, ddk3 = multi.split_datasets()

    # Initialize the flow.
    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.manager)

    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp)
    flow.register_work(bands_work)
    flow.exclude_from_benchmark(bands_work)

    ddk_work = flowtk.Work()
    for inp in [ddk1, ddk2, ddk3]:
        ddk_work.register_ddk_task(inp, deps={bands_work.nscf_task: "WFK"})

    flow.register_work(ddk_work)
    flow.exclude_from_benchmark(ddk_work)

    return flow


def build_flow(options):
    flow = make_base_flow(options)

    optic_input = abilab.OpticInput(
        broadening=0.002,
        domega=0.0003,
        maxomega=0.3,
        scissor=0.000,
        tolerance=0.002,
        num_lin_comp=1,
        lin_comp=11,
        num_nonlin_comp=2,
        nonlin_comp=(123, 222),
    )

    mpi_list = options.mpi_list
    if mpi_list is None:
        mpi_list = [1,2,4,8]
        print("Using mpi_list:", mpi_list)
    else:
        print("Using mpi_list from cmd line:", mpi_list)

    work = flowtk.Work()
    for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
        if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        optic_task = flowtk.OpticTask(optic_input, manager=manager, nscf_node=flow[0].nscf_task, ddk_nodes=flow[1])
        work.register_task(optic_task)

    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
from __future__ import annotations

import os
import tempfile
import warnings


def profile(statement, global_vars, local_vars):
    """
    Run statement under profiler, supplying your own globals and locals

    Example::

        stats = profile("main()", global_vars=globals(), local_vars=locals())
    """
    import pstats
    import cProfile
    _, filename = tempfile.mkstemp()
    cProfile.runctx(statement, global_vars, local_vars, filename=filename)

    s = pstats.Stats(filename)
    s.strip_dirs().sort_stats("time")
    s.print_stats()
    os.remove(filename)
    return s


class HtmlDiff:
    """
    This object produces diff files in HTML format and displays them in the browser.

    Usage example:

    .. code-block:: python

        HtmlDiff(filepaths).open_browser()
    """
    def __init__(self, filepaths: list[str]):
        if len(filepaths) < 2:
            raise ValueError("You need more than one file to compare!")
        self.filepaths = filepaths

    def open_browser(self, diffmode="difflib", **kwargs):
        """
        Generate diff with ``diffmode``, open browser, return exit code.
        """
        try:
            func = getattr(self, diffmode)
        except AttributeError:
            raise AttributeError("Unsupported value for diffmode: `%s`" % str(diffmode))

        return func(**kwargs)

    def _launch_browser(self, tmpname):  # pragma: no cover
        """Open ``tmpname`` file in the default browser."""
        # warning: This code is not portable since we should pass a url.
        if not tmpname.startswith("file://"): tmpname = "file://" + tmpname
        import webbrowser
        try:
            return int(webbrowser.open(tmpname))
        except webbrowser.Error as exc:
            # Warn the user and ignore the exception.
            warnings.warn(str(exc))
            return 1

    def difflib(self, **kwargs):
        """
        Use difflib to generate a HTML file with the diff.
        Open file in the browser.
        """
        with open(self.filepaths[0], 'rt') as fh:
            fromlines = fh.readlines()

        diffs = []
        for path in self.filepaths[1:]:
            with open(path, 'rt') as fh:
                tolines = fh.readlines()

            _, tmpname = tempfile.mkstemp(suffix=".html", text=True)
            import difflib
            #diff = difflib.HtmlDiff().make_table(fromlines, tolines,
            diff = difflib.HtmlDiff().make_file(fromlines, tolines,
                                                self.filepaths[0], path) #context=options.c, numlines=n)
            with open(tmpname, "wt") as fh:
                fh.writelines(diff)

            return self._launch_browser(tmpname)

    def pygmentize(self):
        """
        Execute ``diff`` and ``pygmentize`` in a subprocess to generate a HTML file with the diff.
        Open file in the browser.
        """
        for file2 in self.filepaths[1:]:
            _, tmpname = tempfile.mkstemp(suffix=".html", text=True)

            # https://stackoverflow.com/questions/641055/diff-to-html-diff2html-program
            #cmd = "/usr/bin/diff -y %s %s | pygmentize -l diff -f html -O full -o %s" % (
            #    self.filepaths[0], file2, tmpname)
            cmd = "diff -U9999999 %s %s | pygmentize -l diff -f html -O full -o %s" % (
                self.filepaths[0], file2, tmpname)

            retcode = os.system(cmd)
            if retcode != 0:
                print("Non-zero exit status returned by: `%s", cmd)
                print("Please install pygmentize with `pip install pygmentize` or `conda install pygmentize`")
                return retcode

            return self._launch_browser(tmpname)


def display_top(snapshot, key_type='lineno', limit=3):
    """
    Profile memory usage in Python.
    Taken from https://stackoverflow.com/questions/552744/how-do-i-profile-memory-usage-in-python

    Example::

        tracemalloc.start()
        main()
        snapshot = tracemalloc.take_snapshot()
        display_top(snapshot)
    """
    import tracemalloc
    import linecache
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)

    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        print("#%s: %s:%s: %.1f KiB"
              % (index, filename, frame.lineno, stat.size / 1024))
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print('    %s' % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))



def get_size(bytes, suffix="B"):
    """
    Scale bytes to its proper format
    e.g:
        1253656 => '1.20MB'
        1253656678 => '1.17GB'
    """
    factor = 1024
    for unit in ["", "K", "M", "G", "T", "P"]:
        if bytes < factor:
            return f"{bytes:.2f}{unit}{suffix}"
        bytes /= factor


def print_hardware_system_info() -> None:
    """
    Taken from <https://thepythoncode.com/article/get-hardware-system-information-python>
    """
    import platform
    uname = platform.uname()
    print("="*40, "System Information", "="*40)
    print(f"System: {uname.system}")
    print(f"Node Name: {uname.node}")
    print(f"Release: {uname.release}")
    print(f"Version: {uname.version}")
    print(f"Machine: {uname.machine}")
    print(f"Processor: {uname.processor}")

    import psutil
    # let's print CPU information
    print("="*40, "CPU Info", "="*40)
    # number of cores
    print("Physical cores:", psutil.cpu_count(logical=False))
    print("Total cores:", psutil.cpu_count(logical=True))
    # CPU frequencies
    cpufreq = psutil.cpu_freq()
    print(f"Max Frequency: {cpufreq.max:.2f}Mhz")
    print(f"Min Frequency: {cpufreq.min:.2f}Mhz")
    print(f"Current Frequency: {cpufreq.current:.2f}Mhz")
    # CPU usage
    print("CPU Usage Per Core:")
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True, interval=1)):
        print(f"Core {i}: {percentage}%")
    print(f"Total CPU Usage: {psutil.cpu_percent()}%")

    # Memory Information
    print("="*40, "Memory Information", "="*40)
    # get the memory details
    svmem = psutil.virtual_memory()
    print(f"Total: {get_size(svmem.total)}")
    print(f"Available: {get_size(svmem.available)}")
    print(f"Used: {get_size(svmem.used)}")
    print(f"Percentage: {svmem.percent}%")
    print("="*20, "SWAP", "="*20)
    # get the swap memory details (if exists)
    swap = psutil.swap_memory()
    print(f"Total: {get_size(swap.total)}")
    print(f"Free: {get_size(swap.free)}")
    print(f"Used: {get_size(swap.used)}")
    print(f"Percentage: {swap.percent}%")

    # Disk Information
    print("="*40, "Disk Information", "="*40)
    print("Partitions and Usage:")
    # get all disk partitions
    partitions = psutil.disk_partitions()
    for partition in partitions:
        print(f"=== Device: {partition.device} ===")
        print(f"  Mountpoint: {partition.mountpoint}")
        print(f"  File system type: {partition.fstype}")
        try:
            partition_usage = psutil.disk_usage(partition.mountpoint)
        except PermissionError:
            # this can be catched due to the disk that
            # isn't ready
            continue
        print(f"  Total Size: {get_size(partition_usage.total)}")
        print(f"  Used: {get_size(partition_usage.used)}")
        print(f"  Free: {get_size(partition_usage.free)}")
        print(f"  Percentage: {partition_usage.percent}%")

    # get IO statistics since boot
    disk_io = psutil.disk_io_counters()
    print(f"Total read: {get_size(disk_io.read_bytes)}")
    print(f"Total write: {get_size(disk_io.write_bytes)}")

    # Network information
    print("="*40, "Network Information", "="*40)
    # get all network interfaces (virtual and physical)
    if_addrs = psutil.net_if_addrs()
    for interface_name, interface_addresses in if_addrs.items():
        for address in interface_addresses:
            print(f"=== Interface: {interface_name} ===")
            if str(address.family) == 'AddressFamily.AF_INET':
                print(f"  IP Address: {address.address}")
                print(f"  Netmask: {address.netmask}")
                print(f"  Broadcast IP: {address.broadcast}")
            elif str(address.family) == 'AddressFamily.AF_PACKET':
                print(f"  MAC Address: {address.address}")
                print(f"  Netmask: {address.netmask}")
                print(f"  Broadcast MAC: {address.broadcast}")
    # get IO statistics since boot
    net_io = psutil.net_io_counters()
    print(f"Total Bytes Sent: {get_size(net_io.bytes_sent)}")
    print(f"Total Bytes Received: {get_size(net_io.bytes_recv)}")



# coding: utf-8
"""
Function1D describes a function of a single variable and provides an easy-to-use API
for performing common tasks such as algebraic operations, integrations, differentiations, plots ...
"""
from __future__ import annotations

import numpy as np

from io import StringIO
from typing import Tuple, Union
from monty.functools import lazy_property
from abipy.tools.typing import Figure
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, add_plotly_fig_kwargs, PlotlyRowColDesc, get_fig_plotly)
from abipy.tools.derivatives import finite_diff
from abipy.tools.numtools import data_from_cplx_mode

__all__ = [
    "Function1D",
]


class Function1D:
    """
    Immutable object representing a real|complex function of real variable.
    """

    @classmethod
    def from_constant(cls, mesh, const) -> Function1D:
        """Build a constant function from the mesh and the scalar ``const``"""
        mesh = np.ascontiguousarray(mesh)
        return cls(mesh, np.ones(mesh.shape) * const)

    def __init__(self, mesh, values):
        """
        Args:
            mesh: array-like object with the real points of the grid.
            values: array-like object with the values of the function (can be complex-valued)
        """
        mesh = np.ascontiguousarray(mesh)
        self._mesh = np.real(mesh)
        self._values = np.ascontiguousarray(values)
        assert len(self.mesh) == len(self.values)

    @property
    def mesh(self) -> np.ndarray:
        """|numpy-array| with the mesh points"""
        return self._mesh

    @property
    def values(self) -> np.ndarray:
        """Values of the functions."""
        return self._values

    def __len__(self) -> int:
        return len(self.mesh)

    def __iter__(self):
        return zip(self.mesh, self.values)

    def __getitem__(self, slice) -> Tuple[float, float]:
        return self.mesh[slice], self.values[slice]

    def __eq__(self, other) -> bool:
        if other is None: return False
        return (self.has_same_mesh(other) and
                np.allclose(self.values, other.values))

    def __ne__(self, other) -> bool:
        return not (self == other)

    def __neg__(self) -> Function1D:
        return self.__class__(self.mesh, -self.values)

    def __pos__(self) -> Function1D:
        return self

    def __abs__(self) -> Function1D:
        return self.__class__(self.mesh, np.abs(self.values))

    def __add__(self, other) -> Function1D:
        cls = self.__class__
        if isinstance(other, cls):
            assert self.has_same_mesh(other)
            return cls(self.mesh, self.values+other.values)
        else:
            return cls(self.mesh, self.values + np.array(other))

    __radd__ = __add__

    def __sub__(self, other) -> Function1D:
        cls = self.__class__
        if isinstance(other, cls):
            assert self.has_same_mesh(other)
            return cls(self.mesh, self.values-other.values)
        else:
            return cls(self.mesh, self.values-np.array(other))

    def __rsub__(self, other) -> Function1D:
        return -self + other

    def __mul__(self, other) -> Function1D:
        cls = self.__class__
        if isinstance(other, cls):
            assert self.has_same_mesh(other)
            return cls(self.mesh, self.values*other.values)
        else:
            return cls(self.mesh, self.values*other)

    __rmul__ = __mul__

    def __truediv__(self, other) -> Function1D:
        cls = self.__class__
        if isinstance(other, cls):
            assert self.has_same_mesh(other)
            return cls(self.mesh, self.values/other.values)
        else:
            return cls(self.mesh, self.values/other)

    __rtruediv__ = __truediv__

    def __pow__(self, other) -> Function1D:
        return self.__class__(self.mesh, self.values**other)

    def _repr_html_(self) -> Figure:
        """Integration with jupyter_ notebooks."""
        return self.plot(show=False)

    @property
    def real(self) -> Function1D:
        """Return new :class:`Function1D` with the real part of self."""
        return self.__class__(self.mesh, self.values.real)

    @property
    def imag(self) -> Function1D:
        """Return new :class:`Function1D` with the imaginary part of self."""
        return self.__class__(self.mesh, self.values.real)

    def conjugate(self) -> Function1D:
        """Return new :class:`Function1D` with the complex conjugate."""
        return self.__class__(self.mesh, self.values.conjugate)

    def abs(self) -> Function1D:
        """Return :class:`Function1D` with the absolute value."""
        return self.__class__(self.mesh, np.abs(self.values))

    @classmethod
    def from_func(cls, func, mesh) -> Function1D:
        """
        Initialize the object from a callable.

        :example:

           Function1D.from_func(np.sin, mesh=np.arange(1,10))
        """
        return cls(mesh, np.vectorize(func)(mesh))

    @classmethod
    def from_file(cls, path, comments="#", delimiter=None, usecols=(0, 1)) -> Function1D:
        """
        Initialize an instance by reading data from path (txt format)
        see also :func:`np.loadtxt`

        Args:
            path: Path to the file containing data.
            comments: The character used to indicate the start of a comment; default: '#'.
            delimiter: str, optional The string used to separate values. By default, this is any whitespace.
            usecols: sequence, optional. Which columns to read, with 0 being the first.
                For example, usecols = (1,4) will extract data from the 2nd, and 5th columns.
        """
        mesh, values = np.loadtxt(path, comments=comments, delimiter=delimiter,
                                  usecols=usecols, unpack=True)
        return cls(mesh, values)

    def to_file(self, path, fmt='%.18e', header='') -> None:
        """
        Save data in a text file. Use format fmr. A header is added at the beginning.
        """
        fmt = "%s %s\n" % (fmt, fmt)
        with open(path, "wt") as fh:
            if header: fh.write(header)
            for x, y in zip(self.mesh, self.values):
                fh.write(fmt % (x, y))

    def __repr__(self) -> str:
        return "%s at %s, size = %d" % (self.__class__.__name__, id(self), len(self))

    def __str__(self) -> str:
        stream = StringIO()
        for x, y in zip(self.mesh, self.values):
            stream.write("%.18e %.18e\n" % (x, y))
        return "".join(stream.getvalue())

    def has_same_mesh(self, other: Function1D) -> bool:
        """True if self and other have the same mesh."""
        if self.h is None and other.h is None:
            # Generic meshes.
            return np.allclose(self.mesh, other.mesh)
        else:
            # Check for linear meshes
            return len(self.mesh) == len(other.mesh) and self.h == other.h

    @property
    def bma(self) -> float:
        """Return b-a. f(x) is defined in [a,b]"""
        return self.mesh[-1] - self.mesh[0]

    @property
    def max(self) -> float:
        """Max of f(x) if f is real, max of :math:`|f(x)|` if complex."""
        if not self.iscomplexobj:
            return self.values.max()
        else:
            # Max of |z|
            return np.max(np.abs(self.values))

    @property
    def min(self) -> float:
        """Min of f(x) if f is real, min of :math:`|f(x)|` if complex."""
        if not self.iscomplexobj:
            return self.values.min()
        else:
            return np.max(np.abs(self.values))

    @property
    def iscomplexobj(self) -> bool:
        """
        Check if values is array of complex numbers.
        The type of the input is checked, not the value. Even if the input
        has an imaginary part equal to zero, `np.iscomplexobj` evaluates to True.
        """
        return np.iscomplexobj(self.values)

    @lazy_property
    def h(self) -> Union[float, None]:
        """The spacing of the mesh. None if mesh is not homogeneous."""
        return self.dx[0] if np.allclose(self.dx[0], self.dx) else None

    @lazy_property
    def dx(self) -> np.ndarray:
        """
        |numpy-array| of len(self) - 1 elements giving the distance between two
        consecutive points of the mesh, i.e. dx[i] = ||x[i+1] - x[i]||.
        """
        dx = np.zeros(len(self)-1)
        for (i, x) in enumerate(self.mesh[:-1]):
            dx[i] = self.mesh[i+1] - x
        return dx

    def find_mesh_index(self, value) -> int:
        """
        Return the index of the first point in the mesh whose value is >= value
        -1 if not found
        """
        for i, x in enumerate(self.mesh):
            if x >= value:
                return i
        return -1

    def finite_diff(self, order: int = 1, acc: int = 4) -> Function1D:
        """
        Compute the derivatives by finite differences.

        Args:
            order: Order of the derivative.
            acc: Accuracy. 4 is fine in many cases.

        Returns:
            :class:`Function1D` instance with the derivative.
        """
        if self.h is None:
            raise ValueError("Finite differences with inhomogeneous meshes are not supported")

        return self.__class__(self.mesh, finite_diff(self.values, self.h, order=order, acc=acc))

    def integral(self, start=0, stop=None) -> Function1D:
        r"""
        Cumulatively integrate y(x) from start to stop using the composite trapezoidal rule.

        Returns:
            :class:`Function1D` with :math:`\int y(x) dx`
        """
        if stop is None: stop = len(self.values) + 1
        x, y = self.mesh[start:stop], self.values[start:stop]
        try :
            from scipy.integrate import cumulative_trapezoid as cumtrapz
        except ImportError:
            from scipy.integrate import cumtrapz

        integ = cumtrapz(y, x=x, initial=0.0)

        return self.__class__(x, integ)

    @lazy_property
    def spline(self):
        """Cubic spline with s=0"""
        from scipy.interpolate import UnivariateSpline
        return UnivariateSpline(self.mesh, self.values, s=0)

    @property
    def spline_roots(self):
        """Zeros of the spline."""
        return self.spline.roots()

    def spline_on_mesh(self, mesh) -> Function1D:
        """Spline the function on the given mesh, returns :class:`Function1D` object."""
        return self.__class__(mesh, self.spline(mesh))

    def spline_derivatives(self, x):
        """Returns all derivatives of the spline at point x."""
        return self.spline.derivatives(x)

    def spline_integral(self, a=None, b=None):
        """
        Returns the definite integral of the spline of f between two given points a and b

        Args:
            a: First point. mesh[0] if a is None
            b: Last point. mesh[-1] if a is None
        """
        a = self.mesh[0] if a is None else a
        b = self.mesh[-1] if b is None else b
        return self.spline.integral(a, b)

    @lazy_property
    def integral_value(self):
        r"""Compute :math:`\int f(x) dx`."""
        return self.integral()[-1][1]

    @lazy_property
    def l1_norm(self) -> float:
        r"""Compute :math:`\int |f(x)| dx`."""
        return abs(self).integral()[-1][1]

    @lazy_property
    def l2_norm(self) -> float:
        r"""Compute :math:`\sqrt{\int |f(x)|^2 dx}`."""
        return np.sqrt((abs(self)**2).integral()[-1][1])

    def fft(self) -> Function1D:
        """Compute the FFT transform (negative sign)."""
        # Compute FFT and frequencies.
        from scipy import fftpack
        n, d = len(self), self.h
        fft_vals = fftpack.fft(self.values, n=n)
        freqs = fftpack.fftfreq(n, d=d)

        # Shift the zero-frequency component to the center of the spectrum.
        fft_vals = fftpack.fftshift(fft_vals)
        freqs = fftpack.fftshift(freqs)

        return self.__class__(freqs, fft_vals)

    def ifft(self, x0=None) -> Function1D:
        r"""Compute the FFT transform :math:`\int e+i`"""
        # Rearrange values in the standard order then perform IFFT.
        from scipy import fftpack
        n, d = len(self), self.h
        fft_vals = fftpack.ifftshift(self.values)
        fft_vals = fftpack.ifft(fft_vals, n=n)

        # Compute the mesh of the IFFT output.
        x0 = 0.0 if x0 is None else x0
        stop = 1./d - 1./(n*d) + x0
        mesh = np.linspace(x0, stop, num=n)

        return self.__class__(mesh, fft_vals)

    #def convolve_with_func1d(self, other):
    #    """Convolve self with other."""
    #    assert self.has_same_mesh(other)
    #    from scipy.signal import convolve
    #    conv = convolve(self.values, other.values, mode="same") * self.h
    #    return self.__class__(self.mesh, conv)

    #def gaussian_convolution(self, width, height=None):
    #    """Convolve data with a Gaussian of standard deviation ``width``."""
    #    from abipy.tools.numtools import gaussian
    #    gvals = gaussian(self.mesh, width, center=self.mesh[len(self)//2], height=height)
    #    return self.convolve_with_func1d(Function1D(self.mesh, gvals))

    #def lorentzian_convolution(self, gamma, height=None):
    #    """Convolve data with a Lorentzian of half-width at half-maximum ``gamma``"""
    #    from abipy.tools.numtools import lorentzian
    #    lvals = lorentzian(self.mesh, gamma, center=self.mesh[len(self)//2], height=height)
    #    return self.convolve_with_func1d(Function1D(self.mesh, lvals))

    #def smooth(self, window_len=11, window="hanning"):
    #    from abipy.tools.numtools import smooth
    #    smooth_vals = smooth(self.values, window_len=window_len, window=window)
    #    return self.__class__(self.mesh, smooth_vals)

    #def real_from_kk(self, with_div=True):
    #    """
    #    Compute the Kramers-Kronig transform of the imaginary part
    #    to get the real part. Assume self represents the Fourier
    #    transform of a response function.

    #    Args:
    #        with_div: True if the divergence should be treated numerically.
    #            If False, the divergence is ignored, results are less accurate
    #            but the calculation is faster.

    #    .. seealso:: <https://en.wikipedia.org/wiki/Kramers%E2%80%93Kronig_relations>
    #    """
    #    from scipy.integrate import cumtrapz, quad
    #    from scipy.interpolate import UnivariateSpline
    #    wmesh = self.mesh
    #    num = np.array(self.values.imag * wmesh, dtype=np.double)

    #    if with_div:
    #        spline = UnivariateSpline(self.mesh, num, s=0)

    #    kk_values = np.empty(len(self))
    #    for i, w in enumerate(wmesh):
    #        den = wmesh**2 - w**2
    #        # Singularity is treated below.
    #        den[i] = 1
    #        f = num / den
    #        f[i] = 0
    #        integ = cumtrapz(f, x=wmesh)
    #        kk_values[i] = integ[-1]

    #        if with_div:
    #            func = lambda x: spline(x) / (x**2 - w**2)
    #            w0 = w - self.h
    #            w1 = w + self.h
    #            y, abserr = quad(func, w0, w1, points=[w])
    #            kk_values[i] += y

    #    return self.__class__(self.mesh, (2 / np.pi) * kk_values)

    #def imag_from_kk(self, with_div=True):
    #    """
    #    Compute the Kramers-Kronig transform of the real part
    #    to get the imaginary part. Assume self represents the Fourier
    #    transform of a response function.

    #    Args:
    #        with_div: True if the divergence should be treated numerically.
    #            If False, the divergence is ignored, results are less accurate
    #            but the calculation is faster.

    #    .. seealso:: <https://en.wikipedia.org/wiki/Kramers%E2%80%93Kronig_relations>
    #    """
    #    from scipy.integrate import cumtrapz, quad
    #    from scipy.interpolate import UnivariateSpline
    #    wmesh = self.mesh
    #    num = np.array(self.values.real, dtype=np.double)

    #    if with_div:
    #        spline = UnivariateSpline(self.mesh, num, s=0)

    #    kk_values = np.empty(len(self))
    #    for i, w in enumerate(wmesh):
    #        den = wmesh**2 - w**2
    #        # Singularity is treated below.
    #        den[i] = 1
    #        f = num / den
    #        f[i] = 0
    #        integ = cumtrapz(f, x=wmesh)
    #        kk_values[i] = integ[-1]

    #        if with_div:
    #            func = lambda x: spline(x) / (x**2 - w**2)
    #            w0 = w - self.h
    #            w1 = w + self.h
    #            y, abserr = quad(func, w0, w1, points=[w])
    #            kk_values[i] += y

    #    return self.__class__(self.mesh, -(2 / np.pi) * wmesh * kk_values)

    def plot_ax(self, ax, exchange_xy=False, normalize=False, xfactor=1, yfactor=1, *args, **kwargs) -> list:
        """
        Helper function to plot self on axis ax.

        Args:
            ax: |matplotlib-Axes|.
            exchange_xy: True to exchange the axis in the plot.
            args: Positional arguments passed to ax.plot
            normalize: Normalize the ydata to 1.
            xfactor, yfactor: xvalues and yvalues are multiplied by this factor before plotting.
            kwargs: Keyword arguments passed to ``matplotlib``. Accepts

        ==============  ===============================================================
        kwargs          Meaning
        ==============  ===============================================================
        cplx_mode       string defining the data to print.
                        Possible choices are (case-insensitive): `re` for the real part
                        "im" for the imaginary part, "abs" for the absolute value.
                        "angle" to display the phase of the complex number in radians.
                        Options can be concatenated with "-"
        ==============  ===============================================================

        Returns:
            List of lines added.
        """
        if self.iscomplexobj:
            cplx_mode = kwargs.pop("cplx_mode", "re-im")
        else:
            cplx_mode = kwargs.pop("cplx_mode", "re")

        lines = []
        for c in cplx_mode.lower().split("-"):
            xx, yy = self.mesh, data_from_cplx_mode(c, self.values)
            if xfactor != 1: xx = xx * xfactor
            if yfactor != 1: yy = yy * yfactor
            if normalize: yy = yy / np.max(yy)

            if exchange_xy:
                xx, yy = yy, xx

            lines.extend(ax.plot(xx, yy, *args, **kwargs))

        return lines

    @add_fig_kwargs
    def plot(self, ax=None, **kwargs) -> Figure:
        """
        Plot the function with matplotlib.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.

        ==============  ===============================================================
        kwargs          Meaning
        ==============  ===============================================================
        exchange_xy     True to exchange x- and y-axis (default: False)
        ==============  ===============================================================

        Returns: |matplotlib-Figure|.
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.grid(True)
        exchange_xy = kwargs.pop("exchange_xy", False)
        self.plot_ax(ax, exchange_xy=exchange_xy, **kwargs)

        return fig

    def plotly_traces(self, fig, rcd=None, exchange_xy=False, xfactor=1, yfactor=1, *args, **kwargs):
        """
        Helper function to plot the function with plotly.

        Args:
            fig: plotly.graph_objects.Figure.
            rcd: PlotlyRowColDesc object used when fig is not None to specify the (row, col) of the subplot in the grid.
            exchange_xy: True to exchange the x and y in the plot.
            xfactor, yfactor: xvalues and yvalues are multiplied by this factor before plotting.
            args: Positional arguments passed to 'plotly.graph_objects.Scatter'
            kwargs: Keyword arguments passed to 'plotly.graph_objects.Scatter'.
                Accepts also AbiPy specific kwargs:

        ==============  ===============================================================
        kwargs          Meaning
        ==============  ===============================================================
        cplx_mode       string defining the data to print.
                        Possible choices are (case-insensitive): `re` for the real part
                        "im" for the imaginary part, "abs" for the absolute value.
                        "angle" to display the phase of the complex number in radians.
                        Options can be concatenated with "-"
        ==============  ===============================================================
        """
        rcd = PlotlyRowColDesc.from_object(rcd)
        ply_row, ply_col = rcd.ply_row, rcd.ply_col

        import plotly.graph_objects as go

        if self.iscomplexobj:
            cplx_mode = kwargs.pop("cplx_mode", "re-im")
        else:
            cplx_mode = kwargs.pop("cplx_mode", "re")

        showlegend = False
        if "name" in kwargs: showlegend = True
        showlegend = kwargs.pop("showlegend", showlegend)

        for c in cplx_mode.lower().split("-"):
            xx, yy = self.mesh, data_from_cplx_mode(c, self.values)
            if xfactor != 1: xx = xx * xfactor
            if yfactor != 1: yy = yy * yfactor

            if exchange_xy:
                xx, yy = yy, xx

            fig.add_trace(go.Scatter(x=xx, y=yy, mode="lines", showlegend=showlegend, *args, **kwargs),
                          row=ply_row, col=ply_col)

    @add_plotly_fig_kwargs
    def plotly(self, exchange_xy=False, fig=None, rcd=None, **kwargs):
        """
        Plot the function with plotly.

        Args:
            exchange_xy: True to exchange x- and y-axis (default: False)
            fig: plotly figure or None if a new figure should be created.
            rcd: PlotlyRowColDesc object used when fig is not None to specify the (row, col)
                of the subplot in the grid.

        Returns: plotly-Figure
        """
        fig, _ = get_fig_plotly(fig=fig)
        rcd = PlotlyRowColDesc.from_object(rcd)
        self.plotly_traces(fig, rcd=rcd, exchange_xy=exchange_xy, **kwargs)

        return fig



#!/usr/bin/env python
r"""
Wavefunction file
=================

This example shows how to analyze the wavefunctions
stored in the WFK.nc file.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the DEN.nc file
ncfile = abiopen(abidata.ref_file("si_nscf_WFK.nc"))
print(ncfile)

# The DEN file has a `Structure` an `ElectronBands` object and wavefunctions.
#print(ncfile.structure)

# To plot the KS eigenvalues.
#ncfile.ebands.plot()

# Extract the wavefunction for the first spin, the first band and k=[0.5, 0, 0]
wave = ncfile.get_wave(spin=0, kpoint=[0.5, 0, 0], band=0)
print(wave)

# This is equivalent to
#wave = ncfile.get_wave(spin=0, kpoint=0, band=0)
# because [0.5, 0, 0] is the first k-point in the WFK file

# To visualize the total charge wih vesta
#visu = wave.visualize_ur2("vesta"); visu()

# To plot the wavefunction along the line connecting
# the first and the second in the structure:
#wave.plot_line(point1=0, point2=1)

#wave.plot_line(point1=0, point2=1, with_krphase=True)

# alternatively, one can define the line in terms of two points
# in fractional coordinates:
wave.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], with_krphase=False, num=400)
wave.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], with_krphase=True, num=400)

# To plot the wavefunction along the lines connect the firt atom in the structure
# and all the neighbors within a sphere of radius 3 Angstrom:
#wave.plot_line_neighbors(site_index=0, radius=3)

ncfile.close()



#!/usr/bin/env python
r"""
e-ph ZPR of band edges with generalized Frohlich model
======================================================

Flow to estimate the zero-point renormalization at the band edges
using the generalized Frohlich model. The flow uses DFPT to compute
the effective masses at the band edges (automatically detected by performing a NSCF run with a k-path),
and uses an external DDB file providing BECS, eps_inf and phonon frequencies at Gamma.
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options):

    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Get structure from DDB file.
    ddb_path = abidata.ref_file("refs/mgo_v8t57/mgo_zpr_t57o_DS3_DDB")
    with abilab.abiopen(ddb_path) as ddb:
        structure = ddb.structure

    # Build SCF input using structure from DDB file.
    pseudos = abidata.pseudos("Ca.psp8", "O.psp8")
    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)

    # Set other input variables. These quantities are system-depedent.
    # Here we use parameters similar to https://docs.abinit.org/tests/v8/Input/t57.in
    scf_input.set_vars(
        nband=12,
        nbdbuf=2,
        diemac=6,
        ecut=30,                # Underconverged ecut.
        #ecut=15,
        nstep=100,
        tolvrs=1e-16,
        kptrlatt=[-2,  2,  2,   # In cartesian coordinates, this grid is simple cubic
                   2, -2,  2,
                   2,  2, -2],
    )

    # Build the flow to detect band edges, compute effective masses and finally obtain an estimate for the ZPR
    # BECS/phonons/eps_inf are taken from ddb_node.
    from abipy.flowtk.effmass_works import FrohlichZPRFlow
    flow = FrohlichZPRFlow.from_scf_input(options.workdir, scf_input, ddb_node=ddb_path, ndivsm=2, tolwfr=1e-14,
                                          manager=options.manager)
    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_frohlich_zpr_from_ddb.py -s
#



#!/usr/bin/env python
r"""
Phonon bands with/without the ASR
=================================

This example shows how to plot a phonon band structure
with and without enforcing the acoustic sum rule (ASR).
Both matplotlib and plotly are supported.

.. important::

    Note that a **manager.yml** configuration file and an abinit installation are required
    to run this script as AbiPy needs to invoke anaddb to compute phonons from the DDB file.
"""

#%%
# Open the DDB file with:

from abipy import abilab
import abipy.data as abidata

filepath = abidata.ref_file("mp-1009129-9x9x10q_ebecs_DDB")
ddb = abilab.abiopen(filepath)

#%%
# The ``ddb.anacompare_asr`` method computes the phonon bands and the DOS by calling anaddb
# with different values of asr and returns a PhononBandsPlotter object:
# To make the computation faster, we use the **advanced** options dipdip -1.
# This option should produce results similar to dipdip 1 yet make sure to test
# the effect of this variable before using it in production.

plotter = ddb.anacompare_asr(asr_list=(0, 2), dipdip=-1)
print(plotter)

#%%
# To plot the bands on the same figure with matplotlib, use:

plotter.combiplot()

#%%
# For the plotly version, use:

plotter.combiplotly()

#%%
# To disable the DOS computation, set ``nqsmall` to 0:

plotter = ddb.anacompare_asr(asr_list=(0, 2), nqsmall=0, ndivsm=10, dipdip=-1)

#%%
# To plot the bands on different subplots with matplotlib, use:

plotter.gridplot()

#%%
# For the plotly version, use:
plotter.gridplotly()

#%%
# Finally, remember to close the file with:

ddb.close()



# coding: utf-8
"""
Objects to read and analyze optical properties stored in the optic.nc file produced by optic executable.
"""
from __future__ import annotations

import numpy as np
import abipy.core.abinit_units as abu

from collections import OrderedDict
from monty.string import marquee, list_strings
from monty.functools import lazy_property
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, data_from_cplx_mode
from abipy.abio.robots import Robot
from abipy.electrons.ebands import ElectronBands, ElectronsReader, RobotWithEbands

ALL_CHIS = OrderedDict([
    ("linopt", {
        "longname": "Dielectric function",
        "rank": 2,
        #"terms":
        #"latex": r"\chi(\omega)"
        }
    ),
    ("shg", {
        "longname": "Second Harmonic Generation",
        "rank": 3,
        "terms": ["shg_inter2w", "shg_inter1w", "shg_intra2w",
                  "shg_intra1w", "shg_intra1wS", "shg_chi2tot",
                  "shg_inter2w_AR", "shg_inter1w_AR", "shg_intra2w_AR",
                  "shg_intra1w_AR", "shg_intra1wS_AR", "shg_chi2tot_AR",
                  "shg_chi2full"],
        }
        #"latex": r"\chi(-2\omega, \omega, \omega)"
    ),
    ("leo", {
        "longname": "Linear Electro-optic",
        "rank": 3,
        "terms": ["leo_chi", "leo_eta", "leo_sigma", "leo_chi2tot"],
        }
        #"latex": r"\chi(-\omega, \omega, 0)"
    )
])

#LEO2_TERMS = OrderedDict([
#    ("leo2_chiw", None),
#    ("leo2_etaw", None),
#    ("leo2_chi2w", None),
#    ("leo2_eta2w", None),
#    ("leo2_sigmaw", None),
#    ("leo2_chi2tot", None),
#])


#####################################
# Helper functions for linear optic #
#####################################

def reflectivity(eps):
    """Reflectivity(w) from vacuum, at normal incidence"""
    return np.sqrt(0.5 * (np.abs(eps) + eps.real))


#def abs_coeff(eps):
#    """absorption coeff (in m-1) = omega Im(eps) / c n(eps)"""
#    if (abs(eps(iw)) + dble(eps(iw)) > zero) then
#       tmpabs = aimag(eps(iw))*ene / sqrt(half*( abs(eps(iw)) + dble(eps(iw)) )) / Sp_Lt / Bohr_meter
#    end if


def kappa(eps):
    """Im(refractive index(w)) aka kappa"""
    return np.sqrt(0.5 * (np.abs(eps) - eps.real))


def n(eps):
    """Re(refractive index(w)) aka n"""
    return np.sqrt(0.5 * (np.abs(eps) + eps.real))


#def eels(eps):
#    np.where(np.abs(eps)
#    return - (1 / eps).imag


LINEPS_WHAT2EFUNC = dict(
    n=n,
    reflectivity=reflectivity,
    kappa=kappa,
    re=lambda eps: eps.real,
    im=lambda eps: eps.imag,
    #abs: lambda: eps: np.abs(eps),
    #angle: lambda: eps: np.angle(eps, deg=False),
    #abs_coeff=abs_coeff
    #eels=lambda: eps /
)


class OpticNcFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This file contains the results produced by optic. Provides methods to plot optical
    properties and susceptibilty tensors. Used by :class:`OpticRobot` to analyze multiple files.

    Usage example:

    .. code-block:: python

        with OpticNcFile("out_optic.nc") as optic:
            optic.ebands.plot()
            optic.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: OpticNcFile
    """

    @classmethod
    def from_file(cls, filepath: str) -> OpticNcFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = OpticReader(filepath)

        # Read optic input variables and info on k-point sampling and store them in self.
        keys = [
            "kptopt",
            # optic input variables
            "broadening", "maxomega", "domega", "scissor", "tolerance", "do_antiresonant", "do_ep_renorm",
        ]
        for key in keys:
            setattr(self, key, self.reader.read_value(key))

        # Number of bands in sum over states
        # This variable was added in Abinit v9.6.2.
        # In previous version nband_sum was assumed to be the max number of bands found in the external files.
        self.nband_sum = self.reader.read_value("nband_sum", default=self.ebands.nband)

    @lazy_property
    def wmesh(self):
        """
        Frequency mesh in eV. Note that the same frequency-mesh is used
        for the different optical properties.
        """
        return self.reader.read_value("wmesh")

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, title="Electronic Bands"))

        app(marquee("Optic calculation", mark="="))
        # Show Optic variables.
        app("broadening: %s [Ha], %.3f (eV)" % (self.broadening, self.broadening * abu.Ha_eV))
        app("scissor: %s [Ha], %.3f (eV)" % (self.scissor, self.scissor * abu.Ha_eV))
        app("tolerance: %s [Ha], %.3f (eV)" % (self.tolerance, self.tolerance * abu.Ha_eV))
        app("maxomega: %s [Ha], %.3f (eV)" % (self.maxomega, self.maxomega * abu.Ha_eV))
        app("domega: %s [Ha], %.3f (eV)" % (self.domega, self.domega * abu.Ha_eV))
        app("do_antiresonant %s, do_ep_renorm %s" % (self.do_antiresonant, self.do_ep_renorm))
        app("Number of temperatures: %d" % self.reader.ntemp)
        app("Number of bands in sum-over-states: %d" % self.nband_sum)

        # Show available tensors and computed components.
        for key, info in ALL_CHIS.items():
            if not self.reader.computed_components[key]: continue
            app("%s components computed: %s" % (
                info["longname"], ", ".join(self.reader.computed_components[key])))

        if verbose > 1:
            app(marquee("Abinit Header", mark="="))
            app(self.hdr.to_string(verbose=verbose))

        return "\n".join(lines)

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def has_linopt(self) -> bool:
        """True if the ncfile contains Second Harmonic Generation tensor."""
        return "linopt" in self.reader.computed_components

    @lazy_property
    def has_shg(self) -> bool:
        """True if the ncfile contains Second Harmonic Generation tensor."""
        return "shg" in self.reader.computed_components

    @lazy_property
    def has_leo(self) -> bool:
        """True if the ncfile contains the Linear Electro-optic tensor"""
        return "leo" in self.reader.computed_components

    #@lazy_property
    #def xc(self):
    #    """:class:`XcFunc object with info on the exchange-correlation functional."""
    #    return self.reader.read_abinit_xcfunc()

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    @staticmethod
    def get_linopt_latex_label(what: str, comp: str) -> str:
        """
        Return latex label for linear-optic quantities. Used in plots.
        """
        return dict(
            n=r"$n_{%s}$" % comp,
            reflectivity=r"$R_{%s}$" % comp,
            kappa=r"$\kappa_{%s}$" % comp,
            re=r"$\Re(\epsilon_{%s})$" % comp,
            im=r"$\Im(\epsilon_{%s})$" % comp,
            #abs=r"$|\epsilon_{%s}|$" % comp,
            #abs_coeff=abs_coeff_{%s}} % comp,
            #eels:r"EELS_{%s}" % comp,
        )[what]

    def get_chi2_latex_label(self, key: str, what: str, comp: str) -> str:
        """
        Build latex label for chi2-related quantities. Used in plots.
        """
        symb = "{%s}" % key.capitalize()
        return dict(
            re=r"$\Re(%s_{%s})$" % (symb, comp),
            im=r"$\Im(%s_{%s})$" % (symb, comp),
            abs=r"$|%s_{%s}|$" % (symb, comp),
        )[what]

    @add_fig_kwargs
    def plot_linear_epsilon(self, components="all", what="im", itemp=0,
                            ax=None, xlims=None, with_xlabel=True, label=None, fontsize=12, **kwargs):
        """
        Plot components of the linear dielectric function.

        Args:
            components: List of cartesian tensor components to plot e.g. ["xx", "xy"].
                "all" if all components available on file should be plotted on the same ax.
            what: quantity to plot. "re" for real part, "im" for imaginary. Accepts also "abs", "angle".
            itemp: Temperature index.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            with_xlabel: True if x-label should be added.
            label: True to add legend label to each curve.
            fontsize: Legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        comp2eps = self.reader.read_lineps(components, itemp=itemp)

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        for comp, eps in comp2eps.items():
            values = LINEPS_WHAT2EFUNC[what](eps)
            # Note: I'm skipping the first point at w=0 because optic does not compute it!
            # The same trick is used in the other plots.
            ax.plot(self.wmesh[1:], values[1:],
                    label=self.get_linopt_latex_label(what, comp) if label is None else label)

        ax.grid(True)
        if with_xlabel: ax.set_xlabel('Photon Energy (eV)')
        set_axlims(ax, xlims, "x")
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_linopt(self, select="all", itemp=0, xlims=None, **kwargs):
        """
        Subplots with all linear optic quantities selected by ``select`` at temperature ``itemp``.

        Args:
            select:
            itemp: Temperature index.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used.

        Returns: |matplotlib-Figure|
        """
        key = "linopt"
        if not self.reader.computed_components[key]: return None
        if select == "all": select = list(LINEPS_WHAT2EFUNC.keys())
        select = list_strings(select)

        nrows, ncols = len(select), 1
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=True)

        components = self.reader.computed_components[key]
        for i, (what, ax) in enumerate(zip(select, ax_mat)):
            self.plot_linear_epsilon(what=what, itemp=itemp, components=components,
                                     ax=ax, xlims=xlims, with_xlabel=(i == len(select) - 1),
                                     show=False)
        return fig

    @add_fig_kwargs
    def plot_chi2(self, key, components="all", what="abs", itemp=0, decompose=False,
                  ax=None, xlims=None, with_xlabel=True, label=None, fontsize=12, **kwargs):
        """
        Low-level function to plot chi2 tensor.

        Args:
            key:
            components: List of cartesian tensor components to plot e.g. ["xxx", "xyz"].
                "all" if all components available on file should be plotted on the same ax.
            what: quantity to plot. "re" for real part, "im" for imaginary, Accepts also "abs", "angle".
            itemp: Temperature index.
            decompose: True to plot individual contributions.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            with_xlabel: True to add x-label.
            label: True to add legend label to each curve.
            fontsize: Legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        if not self.reader.computed_components[key]: return None
        comp2terms = self.reader.read_tensor3_terms(key, components, itemp=itemp)

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        for comp, terms in comp2terms.items():
            for name, values in terms.items():
                if not decompose and not name.endswith("tot"): continue
                values = data_from_cplx_mode(what, values)
                ax.plot(self.wmesh[1:], values[1:],
                        label=self.get_chi2_latex_label(key, what, comp) if label is None else label,
                )

        ax.grid(True)
        if with_xlabel: ax.set_xlabel('Photon Energy (eV)')
        set_axlims(ax, xlims, "x")
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_shg(self, **kwargs):
        """Plot Second Harmonic Generation. See plot_chi2 for args supported."""
        return self.plot_chi2(key="shg", show=False, **kwargs)

    @add_fig_kwargs
    def plot_leo(self, **kwargs):
        """Plot Linear Electro-optic. See plot_chi2 for args supported."""
        return self.plot_chi2(key="leo", show=False, **kwargs)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        if self.has_linopt:
            yield self.plot_linear_epsilon(what="re", show=False)
            yield self.plot_linear_epsilon(what="im", show=False)
            yield self.plot_linopt(show=False)
        if self.has_shg:
            yield self.plot_shg(show=False)
        if self.has_leo:
            yield self.plot_leo(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("optic = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(optic)"),
            nbv.new_code_cell("optic.ebands.plot();"),
        ])

        # Add plot calls if quantities have been computed.
        for key, info in ALL_CHIS.items():
            if not self.reader.computed_components[key]: continue
            pycall = "optic.plot_%s();" % key
            nb.cells.extend([
                nbv.new_code_cell(pycall),
            ])

        return self._write_nb_nbpath(nb, nbpath)


class OpticReader(ElectronsReader):
    """
    This object reads the results stored in the optic.nc file
    It provides helper function to access the most important quantities.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: OpticReader
    """
    def __init__(self, filepath: str) -> None:
        super().__init__(filepath)
        self.ntemp = self.read_dimvalue("ntemp")

        self.computed_components = OrderedDict()
        self.computed_ids = OrderedDict()
        for chiname, info in ALL_CHIS.items():
            comp_name = chiname + "_components"
            if comp_name not in self.rootgrp.variables:
                fi_comps, ids = [], []
            else:
                fi_comps = [str(i) for i in self.read_value(comp_name)]
                if info["rank"] == 2:
                    ids = [(int(i[0])-1, int(i[1])-1) for i in fi_comps]
                elif info["rank"] == 3:
                    ids = [(int(i[0])-1, int(i[1])-1, int(i[2])-1) for i in fi_comps]
                else:
                    raise NotImplementedError("rank %s" % info["rank"])

            self.computed_ids[chiname] = ids
            self.computed_components[chiname] = [abu.itup2s(it) for it in ids]

    def read_lineps(self, components, itemp=0):
        """
        Args:
            components: List of cartesian tensor components to plot e.g. ["xx", "xy"].
                "all" if all components available on file should be plotted on the same ax.
            itemp: Temperature index.
        """
        # linopt_epsilon has *Fortran* shape [two, nomega, num_comp, ntemp]
        key = "linopt"
        if components == "all": components = self.computed_components[key]
        if not (self.ntemp > itemp >= 0):
            raise ValueError("Invalid itemp: %s, ntemp: %s" % (itemp, self.ntemp))

        var = self.read_variable("linopt_epsilon")
        od = OrderedDict()
        for comp in list_strings(components):
            try:
                ijp = self.computed_components[key].index(comp)
            except ValueError:
                raise ValueError("epsilon_component %s was not computed" % comp)

            values = var[itemp, ijp]
            od[comp] = values[:, 0] + 1j * values[:, 1]
        return od

    def read_tensor3_terms(self, key, components, itemp=0):
        """
        Args:
            key: Name of the netcdf variable to read.
            components: List of cartesian tensor components to plot e.g. ["xxx", "xyz"].
                "all" if all components available on file should be plotted on the same ax.
            itemp: Temperature index.

        Return:
            :class:`OrderedDict` mapping cartesian components e.g. "xyz" to data dictionary.
            Individual entries are listed in ALL_CHIS[key]["terms"]
        """
        # arrays have Fortran shape [two, nomega, num_comp, ntemp]
        if components == "all": components = self.computed_components[key]
        components = list_strings(components)
        if not (self.ntemp > itemp >= 0):
            raise ValueError("Invalid itemp: %s, ntemp: %s" % (itemp, self.ntemp))

        od = OrderedDict([(comp, OrderedDict()) for comp in components])
        for chiname in ALL_CHIS[key]["terms"]:
            #print("About to read:", chiname)

            try:
                var = self.read_variable(chiname)
            except self.Error as exc:
                # This to support new terms added ina shg
                if key != "shg":
                    raise exc

            for comp in components:
                try:
                    ijkp = self.computed_components[key].index(comp)
                except ValueError:
                    raise ValueError("%s component %s was not computed" % (key, comp))
                values = var[itemp, ijkp]
                od[comp][chiname] = values[:, 0] + 1j * values[:, 1]
        return od


class OpticRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple optic.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: OpticRobot
    """
    EXT = "OPTIC"

    @lazy_property
    def computed_components_intersection(self):
        """
        Dictionary with the list of cartesian tensor components
        available in each file. Use keys from ALL_CHIS.
        """
        od = OrderedDict()
        for ncfile in self.abifiles:
            for chiname in ALL_CHIS:
                comps = ncfile.reader.computed_components[chiname]
                if chiname not in od:
                    od[chiname] = comps
                else:
                    # Build intersection while preserving order.
                    od[chiname] = self.ordered_intersection(od[chiname], comps)
        return od

    @add_fig_kwargs
    def plot_linopt_convergence(self, components="all", what_list=("re", "im"),
                                sortby="nkpt", itemp=0, xlims=None, **kwargs):
        """
        Plot the convergence of the dielectric function tensor with respect to
        parameter defined by ``sortby``.

        Args:
            components: List of cartesian tensor components to plot e.g. ["xx", "xy"].
                "all" if all components available on file should be plotted on the same ax.
            what_list: List of quantities to plot. "re" for real part, "im" for imaginary.
                Accepts also "abs", "angle".
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function.
                If None, no sorting is performed.
                If string, it's assumed that the ncfile has an attribute with the same name and getattr is invoked.
                If callable, the output of callable(ncfile) is used.
            itemp: Temperature index.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.

        Returns: |matplotlib-Figure|
        """
        # Build grid plot: computed tensors along the rows, what_list along columns.
        key = "linopt"
        components = self.computed_components_intersection[key]

        nrows, ncols = len(components), len(what_list)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        label_ncfile_param = self.sortby(sortby)
        for i, comp in enumerate(components):
            for j, what in enumerate(what_list):
                ax = ax_mat[i, j]
                for ifile, (label, ncfile, param) in enumerate(label_ncfile_param):

                    ncfile.plot_linear_epsilon(components=comp, what=what, itemp=itemp, ax=ax,
                        xlims=xlims, with_xlabel=(i == len(components) - 1),
                        label="%s %s" % (sortby, param) if not callable(sortby) else str(param),
                        show=False)

                    if ifile == 0:
                        ax.set_title(ncfile.get_linopt_latex_label(what, comp))

                if (i, j) != (0, 0):
                    ax.legend().set_visible(False)

        return fig

    @add_fig_kwargs
    def plot_shg_convergence(self, **kwargs):
        """Plot Second Harmonic Generation. See plot_convergence_rank3 for args supported."""
        if "what_list" not in kwargs: kwargs["what_list"] = ["abs"]
        return self.plot_convergence_rank3(key="shg", **kwargs)

    @add_fig_kwargs
    def plot_leo_convergence(self, **kwargs):
        """Plot Linear electron-optic. See plot_convergence_rank3 for args supported."""
        if "what_list" not in kwargs: kwargs["what_list"] = ["abs"]
        return self.plot_convergence_rank3(key="leo", **kwargs)

    @add_fig_kwargs
    def plot_convergence_rank3(self, key, components="all", itemp=0, what_list=("abs",),
                               sortby="nkpt", decompose=False, xlims=None, **kwargs):
        """
        Plot convergence of arbitrary rank3 tensor. This is a low-level routine used in other plot methods.

        Args:
            key: Name of the quantity to analyze.
            components: List of cartesian tensor components to plot e.g. ["xxx", "xyz"].
                "all" if all components available on file should be plotted on the same ax.
            itemp: Temperature index.
            what_list: List of quantities to plot. "re" for real part, "im" for imaginary.
                Accepts also "abs", "angle".
            sortby: Define the convergence parameter, sort files and produce plot labels. Can be None, string or function.
                If None, no sorting is performed.
                If string, it's assumed that the ncfile has an attribute with the same name and ``getattr`` is invoked.
                If callable, the output of callable(ncfile) is used.
            decompose: True to plot individual contributions.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used.

        Returns: |matplotlib-Figure|
        """
        # Build grid plot: computed tensors along the rows, what_list along columns.
        components = self.computed_components_intersection[key]

        nrows, ncols = len(components), len(what_list)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        label_ncfile_param = self.sortby(sortby)
        for i, comp in enumerate(components):
            for j, what in enumerate(what_list):
                ax = ax_mat[i, j]
                for ifile, (label, ncfile, param) in enumerate(label_ncfile_param):

                    ncfile.plot_chi2(key=key, components=comp, what=what, itemp=itemp, decompose=decompose,
                        ax=ax, xlims=xlims, with_xlabel=(i == len(components) - 1),
                        label="%s %s" % (sortby, param) if not callable(sortby) else str(param),
                        show=False, **kwargs)

                    if ifile == 0:
                        ax.set_title(ncfile.get_chi2_latex_label(key, what, comp))

                if (i, j) != (0, 0):
                    ax.legend().set_visible(False)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        for key, comps in self.computed_components_intersection.items():
            if not comps: continue
            plot_fig = getattr(self, "plot_%s_convergence" % key)
            yield plot_fig(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to nbpath. If ``nbpath`` is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.OpticRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
        ])

        for key, comps in self.computed_components_intersection.items():
            if not comps: continue
            pycall = "robot.plot_%s_convergence();" % key
            nb.cells.extend([
                nbv.new_code_cell(pycall),
            ])

        # Mixins
        nb.cells.extend(self.get_baserobot_code_cells())
        nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)
